# -*- coding: utf-8 -*-

#####################
## Command Sequence of Quesi Elastic Neutron Scattering
#####################

# special reserved word for commands history
HISTORY = None
# reserved words for return value in this.
DAT = None
ECA = None
EC = None
ecaRes = None
ecaDet = None

import Manyo as mm
#import DNA.ana.Reduction.ElasticScanBase as ES
import utsusemi.DNA.ana.Reduction.QensBase as QB
import utsusemi.DNA.ana.Reduction.PeakFitBase as PF
import utsusemi.DNA.ana.Reduction.SearchPathDNA as SPD

#basepath="/usr/local/mlf/DNA/tmp/"
#search_dir="/home/dnaadmin/ana/setup/"
search_dir=SPD.CheckForOutput("setup")
if search_dir=="":
    raise UserWarning("ElasticScan.py >> Not found setup directory.")

NumOfItteration = 1000

plotflug = False

#########################################
#def GetDataOfDNA(runNo=314,HwParam=" ",LambParam=" ",QParam=" ",TimeParam=" ",initFile="GetQEDataOfDNA.ini",outputData="QE",fileTag="_"):
#    """
#    Load EventData and convert to QE map
#    - GetHistogramHW(runNo,dHW,hw_min,hw_max,tof_offset,startTime,endTime)
#    - DoMask(DAT)
#    - ILambdaCorrDNA(dat,ec)
#    - SolidAngleCorrDNA(dat)
#    - CreateQEMap(dat,startQ,endQ,deltaQ)
#
#    @param runNo        (int) Run Number
#    @param HwParam      (string)
#    @param LambParam    (string)
#    @param QParam       (string)
#    @param TimeParam    (string)
#    @param initFile     (string)
#    @param fileTag      (string)   
#    @retval ECM      (ElementContainerMatrix)
#    """
#
#    return ES.GetDataOfDNA(runNo, HwParam, LambParam, QParam, TimeParam, initFile, outputData, fileTag,search_dir)
#
#########################################

def SliceQE(dat=DAT,lowerQ=0.1,upperQ=0.5):
    """
    最小値、最大値を入力して、その範囲に含まれるECを返す。
    @param dat       (ElementContaienrArray)
    @param lowerQ    (double) lowerQ
    @param upperQ    (double) upperQ
    @retval ec       (ElementContainer)
    """
    return QB.sliceQE(dat,lowerQ,upperQ)

##########################################

def SetEbinForECA(eca=DAT,initial=-0.1,final=0.1,delta=0.01):
    """
    最小値、最大値、刻み幅を入力してECAをEビニングを行う。
    @param  dat        (ElementContaienrArray)
    @param  initial    (double)
    @param  final      (double)
    @param  delta      (double)
    @retval DAT        (ElementContaienrArray)
    """
    return QB.setEbinForECA(eca,initial,final,delta)

##########################################

def SetEbinForEC(ec=EC,initial=-0.1,final=0.1,delta=0.01):
    """
    最小値、最大値、刻み幅を入力してECをEビニングを行う。
    @param  ec         (ElementContaienr)
    @param  initial    (double)
    @param  final      (double)
    @param  delta      (double)
    @retval EC         (ElementContaienr)
    """
    return QB.setEbinForEC(ec,initial,final,delta)

##########################################

def SetEbin(ec=EC,initial=-0.1,final=0.1,delta=0.01):
    """
    最小値、最大値、刻み幅を入力してEビニングを行う。
    @param  ec         (ElementContaienr/Array)
    @param  initial    (double)
    @param  final      (double)
    @param  delta      (double)
    @retval EC         (ElementContaienr/Array)
    """
    return QB.setEbin(ec,initial,final,delta)

##########################################
#
#def SetEbinForECA(eca=DAT,initial=-0.1,final=0.1):
#    """
#    SetEbin
#
#    @param  dat        (ElementContaienrArray)
#    @param  initial    (double)
#    @param  final      (double)
#    @retval DAT        (ElementContaienrArray)
#    """
#    return QB.setEbinForECA(eca,initial,final)
#
##########################################
#
#def SetEbinForEC(ec=EC,initial=-0.1,final=0.1):
#    """
#    SetEbin
#
#    @param  ec         (ElementContaienr)
#    @param  initial    (double)
#    @param  final      (double)
#    @retval EC         (ElementContaienr)
#    """
#    return QB.setEbinForEC(ec,initial,final)
#
##########################################
#
#def SetEbin(ec=EC,initial=-0.1,final=0.1):
#    """
#    SetEbin
#
#    @param  ec         (ElementContaienr/Array)
#    @param  initial    (double)
#    @param  final      (double)
#    @retval EC         (ElementContaienr/Array)
#    """
#    return QB.setEbin(ec,initial,final)
#
##########################################

def SetQbin(dat=ECA,initialQ=0.0,finalQ=1.0,deltaQ=0.1):
    """
    最小値、最大値、刻み幅を入力して、Qビニングを行う。
    @param dat    (ElementContaienrArray)
    @param initialQ  (double)
    @param finalQ    (double)
    @param deltaQ    (double)
    @retval DAT   (ElementContaienrArray)
    """
    return QB.setQbin(dat,initialQ,finalQ,deltaQ)

##########################################

def SetQbinbyText(dat=ECA,Q_list_string=" "):
    """
    カンマ区切りテキストのテーブルを新しいQの境界として、Qビニングを行う。
    @param dat    (ElementContaienrArray)
    @param Q_list_string         (string)
    @retval DAT   (ElementContaienrArray)
    """
    return QB.setQbinbyText(dat,Q_list_string)

##########################################

def SelectFittingQbin(dat=ECA,fitlist=" "):
    """
    カンマ区切り文字列で指定したBoolにより、Fittingを行うかどうかを書き込む。
    Select Fitting Q bin
    -filtlist is comma-separated-string, t or f
    @param dat        (ElementContaienrArray)
    @param fitlist    (string)
    @retval None
    """
    return QB.selectFittingQbin(dat,fitlist)

##########################################

def CheckQbin(dat=ECA):
    """
    Fittingを行うかどうかのCheck。標準出力へ書き込む。
    Check Qbin information
     --retval is Q histogram of EC integrated toward Hw axis

    @param dat    (ElementContaienrArray)
    @retval EC    (ElementContaienr)
    """
    return QB.CheckQbin(dat)

#########################################

def GetIntegralHWData(dat=ECA,emin=0.0,emax=1.0):
    """
    I(Q,E)を入力し、入力した範囲にE方向に積分したI(Q)の数値リストを返す。
    @param dat    (ElementContaienrArray)
    @param emiin   (float)
    @param emax    (float)
    @retval ret    (list)
    """
    return QB.getIntegralHWData(dat,emin,emax)

#########################################

def SetFunction(function="lc dc"):
    """
    フィッティング関数の設定
    Set Fitting Function
    -function is space-separated-stirng
    -function:
     -- Lorentz function = l
     -- Gauss function   = g
     -- convoluted Lorentz function = lc
     -- convoluted Gauss function = gc
     -- convoluted delta function = dc

    @param    function (string)
    @retval   function (string)
    """
    return PF.SetFunction(function)

##########################################

def SetInitialParam(initParam=""):
    """
    フィッティング初期パラメータの設定
    Set Initial fitting parameter
    -initParam is comma-separaed-string

    @param    initParam (string)
    @retval   initParam (string)
    """
    return PF.SetInitialParam(initParam)

##########################################

def LoadInitialParam(filename="test.txt"):
    """
    For Utsusemi interface
    - Load Initial Parameter from external text file

    @param    filename
    @retval   initParam (string)
    """
    param,param_ub,parama_lb = QB.LoadFittingParam(filename)
    return param

##########################################

def LoadInitialLBParam(filename="test.txt"):
    """
    For Utsusemi interface
    - Load Initial Parameter(Lower bound) from external text file

    @param    filename
    @retval   initParam (string)
    """
    param,param_ub,param_lb = QB.LoadFittingParam(filename)
    return param_lb

##########################################

def LoadInitialUBParam(filename="test.txt"):
    """
    For Utsusemi interface
    - Load Initial Parameter(Upper bound) from external text file

    @param    filename
    @retval   initParam (string)
    """
    param,param_ub,parama_lb = QB.LoadFittingParam(filename)
    return param_ub

##########################################

def LoadFittingParam(filename="test.txt"):
    """
    - Load Initial Parameter from external text file

    @param    filename
    @retval   initParam (string)
    """
    return QB.LoadFittingParam(filename)

##########################################

def SaveFittingParam(function,param,param_ub,param_lb,link,filename="test.txt"):
    """
    Save Fitting initial Param

    @param    param(str/list)
    @param    param_ub(str/list)
    @param    param_lb(str/list)
    @param    link(str/list)
    @param    filename(str)
    @retval   None
    """
    return QB.saveFittingParam(function,param,param_ub,param_lb,link,filename)

##########################################

def SetResolutionData(ec=EC):
    """
    Set Resolution function by EC, for convolution fitting

    @param    ec (ElementContainer)
    @retval   None
    """
    return PF.SetResolutionData(ec)

##########################################

def PeakConvFitEC(ecDet,ecRes,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Convolution Fitting script by using Levmar

    @param ecDet           (ElementContainer)
    @param ecRes           (ElementContainer)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval eca            (ElementContainerArray)
    """
    return QB.peakConvFitEC(ecDet,ecRes,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def PeakConvFitECA(ecaDet,ecaRes,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Convolution Fitting script by using Levmar

    @param ecaDet           (ElementContainerArray)
    @param ecaRes           (ElementContainerArray)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval ecm            (ElementContainerMatrix)
    """
    return QB.peakConvFitECA(ecaDet,ecaRes,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def PeakConvFit(ecaDet,ecaRes,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Convolution Fitting script by using Levmar

    @param ecaDet           (ElementContainerArray)
    @param ecaRes           (ElementContainerArray)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval ecm            (ElementContainerMatrix)
    """
    return QB.peakConvFit(ecaDet,ecaRes,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def PeakFitEC(ecDet,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Fitting script by using Levmar

    @param ecDet           (ElementContainer)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval eca            (ElementContainerArray)
    """
    return QB.peakFitEC(ecDet,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def PeakFitECA(ecaDet,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Fitting script by using Levmar

    @param ecaDet           (ElementContainerArray)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval ecm            (ElementContainerMatrix)
    """
    return QB.peakFitECA(ecaDet,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def PeakFit(ecaDet,function,param,param_ub,param_lb,link_value,fit_region="",useDomain=True):
    """
    Peak Fitting script by using Levmar

    @param ecaDet           (ElementContainer/Array)
    @param function        (string)
    @param param           (string)
    @param param_lb        (string)
    @param param_ub        (string)
    @param link_value      (string)
    @param useDomain       (bool)
    @retval ecm            (ElementContainerArray/Matrix)
    """
    return QB.peakFit(ecaDet,function,param,param_ub,param_lb,link_value,fit_region,useDomain,NumOfItteration,plotflug)

##########################################

def GetConvFitGuessEC(ecDet,ecRes,function,param,fit_region=""):
    """
    Get convolution fitting Initial parameter function and obseved data

    @param ecDet           (ElementContainer)
    @param ecRes           (ElementContainer)
    @param function        (string)
    @param param           (string)
    @retval eca            (ElementContainerArray)
    """
    return QB.getConvFitGuessEC(ecDet,ecRes,function,param,fit_region,plotflug)

##########################################

def GetConvFitGuessECA(ecaDet,ecaRes,function,param,fit_region=""):
    """
    Get convolution fitting Initial parameter function and obseved data

    @param ecaDet           (ElementContainerArray)
    @param ecaRes           (ElementContainerArray)
    @param function        (string)
    @param param           (string)
    @retval ecm            (ElementContainerMatrix)
    """
    return QB.getConvFitGuessECA(ecaDet,ecaRes,function,param,fit_region,plotflug)


##########################################

def GetConvFitGuess(ecaDet,ecaRes,function,param,fit_region=""):
    """
    Get convolution fitting Initial parameter function and obseved data

    @param ecaDet           (ElementContainer/Array)
    @param ecaRes           (ElementContainer/Array)
    @param function        (string)
    @param param           (string)
    @retval ecm            (ElementContainerArray/Matrix)
    """
    return QB.getConvFitGuess(ecaDet,ecaRes,function,param,fit_region,plotflug)

##########################################

def GetFitGuessEC(ecDet,function,param,fit_region=""):
    """
    Get fitting Initial parameter function and obseved data

    @param ecDet           (ElementContainer)
    @param function        (string)
    @param param           (string)
    @retval eca            (ElementContainerArray)
    """
    return QB.getFitGuessEC(ecDet,function,param,fit_region,plotflug)

##########################################

def GetFitGuessECA(ecaDet,function,param,fit_region=""):
    """
    Get fitting Initial parameter function and obseved data

    @param ecaDet           (ElementContainerArray)
    @param function        (string)
    @param param           (string)
    @retval ecm            (ElementContainerMatrix)
    """
    return QB.getFitGuessECA(ecaDet,function,param,fit_region,plotflug)

##########################################

def GetFitGuess(ecaDet,function,param,fit_region=""):
    """
    Get fitting Initial parameter function and obseved data

    @param ecaDet           (ElementContainer/Array)
    @param function        (string)
    @param param           (string)
    @retval ecm            (ElementContainerArray/Matrix)
    """
    return QB.getFitGuess(ecaDet,function,param,fit_region,plotflug)

##########################################

def GetResultFunction(result_ecm=DAT,Qnum=0):
    """
    Get resolution(entered) function from Result retval

    @param    result_ecm (ElementContainerMatrix/ElementContainerArray)
    @param    Qnum (int)
    @retval   ec (ElementContainer)
    """
    if type(result_ecm) == mm.ElementContainerMatrix:
        return result_ecm(0)(Qnum)
    if type(result_ecm) == mm.ElementContainerArray:
        return result_ecm(0)

##########################################

def GetObservedData(result_ecm=DAT,Qnum=0):
    """
    Get observed(entered) data from Result retval

    @param    result_ecm (ElementContainerMatrix/ElementContainerArray)
    @param    Qnum (int)
    @retval   ec (ElementContainer)
    """
    if type(result_ecm) == mm.ElementContainerMatrix:
        return result_ecm(1)(Qnum)
    if type(result_ecm) == mm.ElementContainerArray:
        return result_ecm(1)

##########################################

def GetInitialFunction(result_ecm=DAT,Qnum=0):
    """
    Get initial fitting function from Result retval

    @param    result_ecm (ElementContainerMatrix/ElementContainerArray)
    @param    Qnum (int)
    @retval   ec (ElementContainer)
    """
    if type(result_ecm) == mm.ElementContainerMatrix:
        return result_ecm(2)(Qnum)
    if type(result_ecm) == mm.ElementContainerArray:
        return result_ecm(2)

##########################################

def GetResidualFunction(result_ecm=DAT,Qnum=0):
    """
    Get residual function [ (fitted - observed)/sigma ] from Result retval

    @param    result_ecm (ElementContainerMatrix/ElementContainerArray)
    @param    Qnum (int)
    @retval   ec (ElementContainer)
    """
    if type(result_ecm) == mm.ElementContainerMatrix:
        return result_ecm(3)(Qnum)
    if type(result_ecm) == mm.ElementContainerArray:
        return result_ecm(3)

##########################################

def GetSigma(result=DAT):
    """
    Get sigma
    sigma = sqrt sum_i [(fitted - observed)**2/sigma_observed] / fitting-data-num
    @param    result (ElementContainerMatrix/ElementContainerArray)
    @retval   val (list.float)
    """
    return QB.getSigma(result)

##########################################

def GetQValue(eca):
    """
    Get Q Value
    @param    eca (ElementContainerArray)
    @retval   val (list)
    """
    return QB.getQValue(eca)

##########################################

def GetFittingParameterEC(result_ecm=DAT,parameter="1",axis="Q"):
    """
    Get Fitting parameter result function for each Q value

    @param result_ecm  (ElementContainerMatrix)
    @param parameter   (int/string)
    @param axis        (string)
    @retval ecOut      (ElementContainer)
    """
    return QB.getFittingParameterEC(result_ecm(0),parameter,axis)

##########################################
#
#def ShowFitResult(result=DAT):
#    """
#    Show Fitting Result, input is fitting result reval
#
#    @param result          (ElementContainerArray/Matrix)
#    """
#    return QB.showFitResult(result)
#
##########################################
#
#def ShowFitResultEC(result=DAT):
#    """
#    Show Fitting Result, input is fitting result reval
#
#    @param result          (ElementContainerArray)
#    """
#    return QB.showFitResultEC(result)
#
##########################################
#
#def ShowFitResultECA(result=DAT):
#    """
#    Show Fitting Result, input is fitting result reval
#
#    @param result          (ElementContainerMatrix)
#    """
#    return QB.showFitResultECA(result)
#
##########################################
#
#def ShowFitGuess(result=DAT):
#    """
#    Show Fitting Guess Result, input is reval of GetFitGuess or GetConvFitGuess
#
#    @param result          (ElementContainerArray/Matrix)
#    """
#    return QB.showFitGuess(result)
#
##########################################
#
#def ShowFitGuessEC(result=DAT):
#    """
#    Show Fitting Guess Result, input is reval of GetFitGuess or GetConvFitGuess
#
#    @param result          (ElementContainerArray)
#    """
#    return QB.showFitGuessEC(result)
#
##########################################
#
#def ShowFitGuessECA(result=DAT):
#    """
#    Show Fitting Guess Result, input is reval of GetFitGuess or GetConvFitGuess
#
#    @param result          (ElementContainerMatrix)
#    """
#    return QB.showFitGuessECA(result)
#
##########################################
#
#def ShowFittingParameterPlot(result=DAT):
#    """
#    Show Fitting parameter Plot, input is reval of GetFittingParameterEC
#
#    @param result          (ElementContainer)
#    """
#    return QB.showFittingParameterPlot(result)
#
##########################################

def GetFitEachResult(result=DAT):
    """
    Get each components of Fitting result function

    @param result          (ElementContainerMatrix)
    """
    return QB.getFitEachResult(result)

##########################################

def ExportCsvFitResult(result=DAT,filename="fitlog.txt"):
    """
    Export Fitting Result to Csv file

    @param result          (ElementContainerArray)
    @param filename        (str)
    """
    return QB.exportCsvFitResult(result,filename)

##########################################

def ExportCsvFitParamResult(result=DAT,filename="fitlog.txt"):
    """
    Export Fitting Param Result to Csv file

    @param result          (ElementContainerArray)
    @param filename        (str)
    """
    return QB.exportCsvFitParamResult(result,filename)

##########################################

def GetFittingParamResult(result=DAT):
    """
    @param result          (ElementContainerArray)
    @param filename        (str)
    @retval (list)
    """
    return QB.getFittingParamResult(result)

##########################################

def SelectQbin(eca,num):
    """
    Select Qbin for Individual Q-bin Fitting

    @param eca          (ElementContainerArray)
    @param num          (int)
    @retval ec          (ElementContainer)
    """
    return QB.selectQbin(eca,num)

##########################################

def GetParamList(function):
    """
    Select Qbin for Individual Q-bin Fitting

    @param function     (str)
    @retval outputText  (list)
    """
    return QB.getParamList(function)

##########################################

def OutputECText(ec, filename="qens.txt"):
    """
    Output EC data to text file

    @param ec                     (ElementContainer)
    @param filename                    (string)
    """
    return QB.OutputECText(ec,filename)

##########################################

def ReadECA(filename=""):
    """
    @param filename  (string)
    @retval eca      (ElementContainerArray)
    """
    return QB.ReadECA(filename)

##########################################

def OutputECA(eca,filepath):
    """
    @param  eca         (ElementContainerArray)
    @param  filepath    (str)
    @retval None
    """
    return QB.OutputECA(eca,filepath)

##########################################

def GetFunctionName(function=""):
    """
    @param function  (string)
    @retval listFuncName  (list)
    """
    return QB.getFunctionName(function)


##########################################

def GetParamName(function=""):
    """
    @param function  (string)

    @retval listParamName  (list)
    """
    return QB.getParamName(function)


##########################################

def ElementContainerToXYList(ec=EC):
    """
    @param ec        (ElementContainer)
    @retval xylist   (list)
    """
    return QB.ElementContainerToXYList(ec)

##########################################

def ElementContainerArrayToXYList2(eca=DAT):
    """
    @param eca        (ElementContainerArray)
    @retval xylist    (list)
    """
    return QB.ElementContainerArrayToXYList2(eca)

##########################################

def GetNumOfEC(eca=DAT):
    """
    @param eca       (ElementContainerArray)
    @retval val      (int)
    """
    return QB.getNumOfEC(eca)

##########################################

def GetUnit(eca=DAT):
    """
    @param eca       (ElementContainerArray)
    @retval val      (list)
    """
    return QB.getUnit(eca)

##########################################

def GetKeyName(eca=DAT):
    """
    @param eca       (ElementContainerArray)
    @retval val      (list)
    """
    return QB.getKeyName(eca)


##########################################

def GetQBoundary(eca):
    """
    @param eca        (ElementContainerArray)
    @retval qlist     (list)
    """
    return QB.getQBoundary(eca)

##########################################

def SetConvolutionRange(eca,rangelist):
    """
    @param eca        (ElementContainerArray)
    @param rangelist  (list)
    @retval eca       (ElementContainerArray)
    """
    return QB.setConvolutionRange(eca,rangelist)

##########################################

def GetBinInformation(eca=DAT):
    """
    @param eca       (ElementContainerArray)
    @retval val      (list)
    """
    return QB.getBinInformation(eca)

##########################################

def GetFittingRawData(eca):
    """
    @param eca        (ElementContainerArray)
    @retval xylist    (list)
    """
    return QB.getFittingRawData(eca)

##########################################

def GetFittingRawDataWithError(eca):
    """
    @param eca        (ElementContainerArray)
    @retval xylist    (list)
    """
    return QB.getFittingRawDataWithError(eca)

##########################################

def GetFittingResultData(ecm=DAT):
    """
    @param ecm        (ElementContainerArray)
    @retval xylist    (list)
    """
    return QB.getFittingResultData(ecm)

##########################################

def GetResidualDistributionData(ecm):
    """
    @param ecm        (ElementContainerMatrix)
    @retval xylist    (list)
    """
    return QB.getResidualDistributionData(ecm)

##########################################

def GetFittingParameterData(result_ecm=DAT,parameter="1",axis="Q"):
    """
    @param ecm        (ElementContainerMatrix)
    @retval xylist    (list)
    """
    return QB.getFittingParameterData(result_ecm(0),parameter,axis)


##########################################

def GetDeltaFuncResolECA(ecaIn):
    """
    @param ecaIn        (ElementContainerArray)
    @retval ecaOut      (ElementContainerArray)
    """
    return QB.getDeltaFuncResolECA(ecaIn)

##########################################

def UsePastExecuteResult(ecmIn,num,resultlist):
    """
    @param ecmIn        (ElementContainerMatrix)
    @param num        (int)
    @param resultlist        (list)
    @retval ecmOut      (ElementContainerMatrix)
    """
    return QB.usePastExecuteResult(ecmIn,num,resultlist)

##########################################

def ChceckZeroYValueForECA(eca):
    """
    @param  eca         (ElementContainerArray)
    @retval ecaNew      (ElementContainerArray)
    """
    return QB.chceckZeroYValueForECA(eca)

##########################################

def ImportDaveFormat(filepath):
    """
    @param  filepath    (str)
    @retval eca         (ElementContainerArray)
    """
    return QB.importDaveFormat(filepath)

##########################################

def ExportDaveFormat(eca,filepath):
    """
    @param  eca         (ElementContainerArray)
    @param  filepath    (str)
    @retval None
    """
    return QB.exportDaveFormat(eca,filepath)

##########################################

def ImportTextFormat(filepath):
    """
    @param  filepath    (str)
    @retval eca         (ElementContainerArray)
    """
    return QB.importTextFormat(filepath)

##########################################

def ExportTextFormat(eca,filepath):
    """
    @param  eca         (ElementContainerArray)
    @param  filepath    (str)
    @retval None
    """
    return QB.exportTextFormat(eca,filepath)

##########################################

# Dictionary for entry the name of functions
_functions = {
                "SliceQE":SliceQE,
                "SetQbin":SetQbin,
                "SetQbinbyText":SetQbinbyText,
                "SetEbin":SetEbin,
                "SelectFittingQbin":SelectFittingQbin,
                "CheckQbin":CheckQbin,
                "SetFunction":SetFunction,
                "SetInitialParam":SetInitialParam,
                "LoadInitialParam":LoadInitialParam,
                "LoadInitialUBParam":LoadInitialUBParam,
                "LoadInitialLBParam":LoadInitialLBParam,
                "PeakConvFit":PeakConvFit,
                "PeakFit":PeakFit,
                "GetConvFitGuess":GetConvFitGuess,
                "GetFitGuess":GetFitGuess,
                "GetResultFunction":GetResultFunction,
                "GetObservedData":GetObservedData,
                "GetResidualFunction":GetResidualFunction,
                "GetInitialFunction":GetInitialFunction,
                "GetSigma":GetSigma,
                "GetFittingParameterEC":GetFittingParameterEC,
                "GetFitEachResult":GetFitEachResult,
                "ExportCsvFitResult":ExportCsvFitResult,
                "SelectQbin":SelectQbin,
                "OutputECText":OutputECText,
                "ReadECA":ReadECA             }
                #"GetDataOfDNA":GetDataOfDNA,
                #"ShowFitResult":ShowFitResult,
                #"ShowFitGuess":ShowFitGuess,
                #"ShowFittingParameterPlot":ShowFittingParameterPlot,




# -*- coding: utf-8 -*-
#
# $Date: 2013-03-06 18:51:19 +0900 (水, 06 3 2013) $
# $Rev: 55 $
#
import vector
import VTKPanel
import define

import math
import decimal

import numpy as np

class QPointBase(object):
    def __init__(self, param):
        self.positoin = None
        self.index = None
        self.param = param
        self.__isVisible = False

        self.ResetConfig()
        return

    def SetPosition(self, qx, qy, qz):
        self.position = (qx, qy, qz)
        return

    def SetIndex(self, i, j, k):
        self.index = (i, j, k)
        return

    def GetIndexString(self):
        return "% d,% d,% d" % (self.index[0], self.index[1], self.index[2])

    def GetEtmax(self):
        etmax = None
        for name in self.param.GetAllRegionNames():
            et = self.status[name][define.QPVAL_ET]
            if etmax == None:
                if et != None:
                    etmax = et
                    pass
                pass
            else:
                if et != None:
                    etmax = max(etmax, et)
                    pass
                pass
            pass
        return etmax

    def ResetConfig(self):
        self.status = dict()

        for name in self.param.GetAllRegionNames():
            self.status[name] = {define.QPVAL_ET: None,
                                 define.QPVAL_INSIDE: False,
                                 define.QPVAL_PXNO: None,
                                 define.QPVAL_PSDID: None}
            pass

        self.numInsidePoints = 0
        return

    def IsInside(self):
        return (self.numInsidePoints > 0)

    def GetInfoText(self):
        info = "Q point Index: " + str(self.index) + "\n"
        info += "Position [1/A]:\n"
        info += "  (%g, %g, %g)\n" % (self.position[0], self.position[1], self.position[2])
        info += "\n"
        info += "Detectable regions:\n"

        flag = False

        for name, vals in self.status.items():
            if vals[define.QPVAL_INSIDE]:
                flag = True
                info += "  " + name + "\n"
                info += "    Energy transfer [meV]: %g" % vals[define.QPVAL_ET] + "\n"
                info += "    PSD id: " + str(vals[define.QPVAL_PSDID]) + "\n"
                info += "    Pixel no.: " + str(vals[define.QPVAL_PXNO]) + "\n"
                pass

        if not flag:
            info += "  None"
            pass
        return info

    def CalcConfig(self, regions):
        """
        This class must be overrided by inheritance class
        """
        raise NotImplementedError

    def SetVisible(self, isVisible):
        self.__isVisible = isVisible
        return

    def IsVisible(self):
        return self.__isVisible

class DNAQPoint(QPointBase):
    def __init__(self, param):
        QPointBase.__init__(self, param)
        return

    def CalcConfig(self, regions):
        self.ResetConfig()

        q = self.position

        for region in regions:
            props = self.param.detectorProps[region]

            ratio = props.y0 / props.RA
            ratio2 = ratio * ratio

            check = (math.pi / props.d_miller) ** 2 - ratio2 * q[1] ** 2

            if check < 0.0:
                continue

            kf = math.sqrt(check / (1.0 - ratio2))

            if abs(q[1] / kf) > 1.0:
                continue

            theta_m = math.asin(q[1] / kf)

            if theta_m < props.theta_m_min or theta_m > props.theta_m_max:
               continue

            check = q[0] / (kf * math.cos(theta_m))
            if abs(check) > 1.0:
                continue

            phi0 = math.asin(check)
            if phi0 >= 0:
                phi1 = math.pi - phi0
                pass
            else:
                phi1 = -math.pi - phi0
                pass

            for phi in [phi0, phi1]:
                if phi < props.phi_s or phi > props.phi_e:
                    continue

                ki = kf * math.cos(theta_m) * math.cos(phi) - q[2]

                if ki < self.param.kimin or ki > self.param.kimax:
                    continue

                PSD = int((props.phi_e - phi) // props.rPSD + props.PSD0)

                self.status[region][define.QPVAL_ET] = define.MEV_ENG_FAC * (ki ** 2 - kf ** 2)
                self.status[region][define.QPVAL_PSDID] = PSD
                self.status[region][define.QPVAL_INSIDE] = True
                self.numInsidePoints += 1

                for Px in xrange(0, props.nPx):
                    if props.theta_m[Px + 1] <= theta_m < props.theta_m[Px]:
                        self.status[region][define.QPVAL_PXNO] = Px
                        break;
                    pass
                pass
            pass
        return

class DGTypeQPoint(QPointBase):
    def __init__(self, param):
        QPointBase.__init__(self, param)
        return

    def CalcConfig(self, regions):
        self.ResetConfig()

        ki = self.param.ki
        q = self.position

        for region in regions:
            props = self.param.detectorProps[region]

            kf = math.sqrt(q[0] ** 2 + q[1] ** 2 + (q[2] + ki) ** 2)
            if kf > self.param.kfmax or kf < self.param.kfmin:
                continue

            phi = math.atan2(q[0], q[2] + ki)

            if phi < props.phi_s or phi > props.phi_e:
                continue

            # Calc. local PSD index
            locPSD = int(math.floor(props.mR - props.L2 / props.dPSD * math.tan(phi - props.rR0)))
            if locPSD < 0 or locPSD > props.nPSD - 1:
                continue

            theta = math.atan2(q[1], math.sqrt(q[0] ** 2 + (q[2] + ki) ** 2))
            L2x = math.sqrt(props.L2 ** 2 + ((props.mR - locPSD) * props.dPSD) ** 2)

            Px = int((props.yd0 - L2x * math.tan(theta)) // props.hPx)
            if Px < 0 or Px > props.nPx - 1:
                continue

            self.status[region][define.QPVAL_ET] = define.MEV_ENG_FAC * (ki ** 2 - kf ** 2)
            self.status[region][define.QPVAL_PSDID] = locPSD + props.PSD0
            self.status[region][define.QPVAL_INSIDE] = True
            self.status[region][define.QPVAL_PXNO] = Px
            self.numInsidePoints += 1
            pass
        return

DEFALUT_AMT_FILEPATH = u'./Default.amt'

NUM_WAVE_NUMBERS = 30

def GetRoundingRange(range, digitNum):
    fac = math.pow(10.0, digitNum)
    return [math.floor(range[0] / fac) * fac,
            math.ceil(range[1] / fac) * fac]

def ApplyMat3(M, x):
    return [M[0][0] * x[0] + M[0][1] * x[1] + M[0][2] * x[2],
            M[1][0] * x[0] + M[1][1] * x[1] + M[1][2] * x[2],
            M[2][0] * x[0] + M[2][1] * x[1] + M[2][2] * x[2]]

class QSpaceManager(object):
    def __init__(self, mode, filename = DEFALUT_AMT_FILEPATH):
        self.__mode = mode

        self.__qPointArray = []
        self.__qPointHash = dict()
        self.status = Status()
        self.status.Read(filename)
        self.ltc = Lattice(self.status)

        self.assistantConf = AssistantConfs(mode)

        elemRegions = self.assistantConf.GetRegions()

        if self.__mode == 'DNA':
            self.param = DNADetectorParams(elemRegions)
            pass
        elif self.__mode == 'SIK':
            self.param = DGTypeDetectorParam(elemRegions)
            pass

        self.__detectorRegions = dict()
        self.__thinDetectorRegions = dict()
        self.__intensityRegions = dict()
        for region in self.param.GetAllRegionNames():
            self.__detectorRegions[region] = dict()
            self.__thinDetectorRegions[region] = None
            self.__intensityRegions[region] = None
            pass

        self.UpdateEnergyRange()
        return

    def UpdateEnergyRange(self):
        elemRegions = self.assistantConf.GetRegions()

        if self.__mode == 'DNA':
            eimin = self.status.GetFloat(TAG_ENG_EIMIN)
            eimax = self.status.GetFloat(TAG_ENG_EIMAX)
            self.param.SetEnergyRange(eimin, eimax)
            pass
        elif self.__mode == 'SIK':
            ei = self.status.GetFloat(TAG_ENG_EI)
            etmin = self.status.GetFloat(TAG_ENG_ETMIN)
            etmax = self.status.GetFloat(TAG_ENG_ETMAX)
            if ei == etmax:
                # Following numeric factor avoids singularity point of detector region,
                #  which causes uncertain normal direction of isosurfaces.
                etmax = 0.99999 * ei
                pass
            self.param.SetEnergyRange(ei, etmin, etmax)
            pass
        return

    def GetBoundIndices(self):
        qxmin = self.__bounds[0]
        qxmax = self.__bounds[1]
        qymin = self.__bounds[2]
        qymax = self.__bounds[3]
        qzmin = self.__bounds[4]
        qzmax = self.__bounds[5]

        A = [[self.ltc.ra_rotate[0], self.ltc.rb_rotate[0], self.ltc.rc_rotate[0]],
             [self.ltc.ra_rotate[1], self.ltc.rb_rotate[1], self.ltc.rc_rotate[1]],
             [self.ltc.ra_rotate[2], self.ltc.rb_rotate[2], self.ltc.rc_rotate[2]]]

        M = [[A[1][1] * A[2][2] - A[1][2] * A[2][1],
              A[0][2] * A[2][1] - A[0][1] * A[2][2],
              A[0][1] * A[1][2] - A[0][2] * A[1][1]],
             [A[1][2] * A[2][0] - A[1][0] * A[2][2],
              A[0][0] * A[2][2] - A[0][2] * A[2][0],
              A[0][2] * A[1][0] - A[0][0] * A[1][2]],
             [A[1][0] * A[2][1] - A[1][1] * A[2][0],
              A[0][1] * A[2][0] - A[0][0] * A[2][1],
              A[0][0] * A[1][1] - A[0][1] * A[1][0]]]

        det = A[0][0] * A[1][1] * A[2][2] + \
            A[1][0] * A[2][1] * A[0][2] + \
            A[2][0] * A[0][1] * A[1][2] - \
            A[0][0] * A[2][1] * A[1][2] - \
            A[1][0] * A[0][1] * A[2][2] - \
            A[2][0] * A[1][1] * A[0][2]

        for i in range(3):
            for j in range(3):
                M[i][j] /= det
                pass
            pass

        phi = [ApplyMat3(M, [qxmin, qymin, qzmin]),
               ApplyMat3(M, [qxmin, qymax, qzmin]),
               ApplyMat3(M, [qxmax, qymin, qzmin]),
               ApplyMat3(M, [qxmax, qymax, qzmin]),
               ApplyMat3(M, [qxmin, qymin, qzmax]),
               ApplyMat3(M, [qxmin, qymax, qzmax]),
               ApplyMat3(M, [qxmax, qymin, qzmax]),
               ApplyMat3(M, [qxmax, qymax, qzmax])]

        imin = phi[0][0]; imax = phi[0][0]
        jmin = phi[0][1]; jmax = phi[0][1]
        kmin = phi[0][2]; kmax = phi[0][2]

        for x in phi:
            imin = min(x[0], imin)
            imax = max(x[0], imax)
            jmin = min(x[1], jmin)
            jmax = max(x[1], jmax)
            kmin = min(x[2], kmin)
            kmax = max(x[2], kmax)
            pass

        imin = int(math.ceil(imin))
        jmin = int(math.ceil(jmin))
        kmin = int(math.ceil(kmin))

        imax = int(math.floor(imax))
        jmax = int(math.floor(jmax))
        kmax = int(math.floor(kmax))
        return [imin, imax, jmin, jmax, kmin, kmax]

    def CreateDetectableRegion(self, mode):
        for name in self.param.GetAllRegionNames():
            props = self.param.detectorProps[name]

            nPx = props.nPx
            nPSD = props.nPSD
            rdpP = props.rdpP

            gridSize = (NUM_WAVE_NUMBERS + 1) * (nPx / rdpP + 1) * (nPSD + 1)

            Pxv = np.linspace(0, nPx, nPx / rdpP + 1)[np.newaxis, :, np.newaxis]
            PSDv = np.linspace(0, nPSD, nPSD + 1)[np.newaxis, np.newaxis, :]

            if mode == 'DNA':
                kiv = np.linspace(self.param.kimin,
                                  self.param.kimax,
                                  NUM_WAVE_NUMBERS + 1)[:, np.newaxis, np.newaxis]

                theta_mv = props.theta_m[::rdpP][np.newaxis, :, np.newaxis]

                zero_PSDv = np.zeros(PSDv.shape)
                zero_kiv = np.zeros(kiv.shape)

                phi = props.phi_e - props.rPSD * PSDv

                sin_phi = np.sin(phi)
                cos_phi = np.cos(phi)

                theta_B = np.arccos(props.y0 / props.RA * np.cos(theta_mv))

                kf = math.pi / (props.d_miller * np.sin(theta_B))

                qz0 = kf * np.cos(theta_mv)
                qy0 = kf * np.sin(theta_mv)

                qx = qz0 * sin_phi
                qy = qy0 + zero_PSDv
                qz = qz0 * cos_phi

                points = np.array([qx + zero_kiv, qy + zero_kiv, qz - kiv])
                etvals = (kiv ** 2 - kf ** 2) * define.MEV_ENG_FAC + zero_PSDv
                pxvals = Pxv + np.zeros(kiv.shape) + np.zeros(PSDv.shape)
                pass
            elif mode == 'SIK':
                kfv = np.linspace(self.param.kfmin,
                                  self.param.kfmax,
                                  NUM_WAVE_NUMBERS + 1)[:, np.newaxis, np.newaxis]

                w = (props.mR - PSDv) * props.dPSD
                phi = props.rR0 + np.arctan2(w, props.L2)
                L2x = np.sqrt(props.L2 ** 2 + w ** 2)

                cos_phi = np.cos(phi)
                sin_phi = np.sin(phi)

                theta_m = np.arctan2(props.yd0 - Pxv * props.hPx, L2x)

                cos_theta_m = np.cos(theta_m)
                sin_theta_m = np.sin(theta_m)

                qz0 = kfv * cos_theta_m
                qy0 = kfv * sin_theta_m

                qx = qz0 * sin_phi
                qy = qy0
                qz = qz0 * cos_phi - self.param.ki

                points = np.array([qx, qy, qz])
                etvals = define.MEV_ENG_FAC * (self.param.ki ** 2 - kfv ** 2) \
                       + np.zeros(Pxv.shape) + np.zeros(PSDv.shape)
                pxvals = Pxv + np.zeros(kfv.shape) + np.zeros(PSDv.shape)
                pass
            else:
                assert 0, "Illegal Detector Parameter"

            self.__detectorRegions[name]['bounds'] = (np.amin(points[0]), np.amax(points[0]),
                                                      np.amin(points[1]), np.amax(points[1]),
                                                      np.amin(points[2]), np.amax(points[2]))

            points = points.transpose().reshape(3 * gridSize)
            etvals = etvals.transpose().reshape(gridSize)
            pxvals = pxvals.transpose().reshape(gridSize)

            self.__detectorRegions[name]['points'] = points
            self.__detectorRegions[name]['etvals'] = etvals
            self.__detectorRegions[name]['pxvals'] = pxvals
            self.__detectorRegions[name]['dimensions'] = ((NUM_WAVE_NUMBERS + 1), (nPx / rdpP + 1), (nPSD + 1))
            pass
        return

    def GetDetectorRegions(self):
        return self.__detectorRegions

    def CreateThinDetectableRegion(self, mode, etmin, etmax):
        thinDetectableRegionPoints = dict()

        for name in self.param.GetAllRegionNames():
            props = self.param.detectorProps[name]

            nPx = props.nPx
            nPSD = props.nPSD
            rdpP = props.rdpP

            thinDetectableRegionPoints[name] = dict()
            thinDetectableRegionPoints[name]['ponits'] = []
            thinDetectableRegionPoints[name]['dimensions'] = (2, nPx / rdpP + 1, nPSD + 1)

            if mode == 'DNA':
                for locPSD in xrange(props.nPSD + 1):
                    phi = props.phi_e - props.rPSD * locPSD

                    sin_phi = math.sin(phi)
                    cos_phi = math.cos(phi)

                    for Px in xrange(0, nPx + 1, rdpP):
                        theta_m = props.theta_m[Px]
                        theta_B = math.acos(props.y0 / props.RA * math.cos(theta_m))

                        lam = 2.0 * props.d_miller * math.sin(theta_B)
                        kf = 2.0 * math.pi / lam

                        qz0 = kf * math.cos(theta_m)
                        qy0 = kf * math.sin(theta_m)

                        qx = qz0 * sin_phi
                        qy = qy0
                        qz = qz0 * cos_phi

                        kimin = math.sqrt(etmin / define.MEV_ENG_FAC + kf ** 2)
                        kimax = math.sqrt(etmax / define.MEV_ENG_FAC + kf ** 2)

                        thinDetectableRegionPoints[name]['ponits'].append((qx, qy, qz - kimin))
                        thinDetectableRegionPoints[name]['ponits'].append((qx, qy, qz - kimax))
                        pass
                    pass
                pass
            elif mode == 'SIK':
                ki = self.param.ki

                for locPSD in xrange(0, nPSD + 1):
                    w = (props.mR - locPSD) * props.dPSD
                    phi = props.rR0 + math.atan2(w, props.L2)
                    L2x = math.sqrt(props.L2 ** 2 + w ** 2)

                    cos_phi = math.cos(phi)
                    sin_phi = math.sin(phi)

                    for Px in xrange(0, nPx + 1, rdpP):
                        theta_m = math.atan2(props.yd0 - Px * props.hPx, L2x)

                        cos_theta_m = math.cos(theta_m)
                        sin_theta_m = math.sin(theta_m)

                        kf = math.sqrt(ki ** 2 - etmin / define.MEV_ENG_FAC)

                        qz0 = kf * cos_theta_m
                        qy0 = kf * sin_theta_m

                        qx = qz0 * sin_phi
                        qy = qy0
                        qz = qz0 * cos_phi

                        thinDetectableRegionPoints[name]['ponits'].append((qx, qy, qz - ki))

                        kf = math.sqrt(ki ** 2 - etmax / define.MEV_ENG_FAC)

                        qz0 = kf * cos_theta_m
                        qy0 = kf * sin_theta_m

                        qx = qz0 * sin_phi
                        qy = qy0
                        qz = qz0 * cos_phi

                        thinDetectableRegionPoints[name]['ponits'].append((qx, qy, qz - ki))
                        pass
                    pass
                pass
            else:
                assert 0, "Illegal Detector Parameter"

            pass
        return thinDetectableRegionPoints

    def SetBoundingBox(self, margin, checkedRegions):
        firstLoop = True
        range = [None, None, None, None, None, None]

        if len(checkedRegions) == 0:
            checkedRegions = self.__detectorRegions.keys()
            pass

        for region in checkedRegions:
            bounds = self.__detectorRegions[region]['bounds']

            for i in (0, 2, 4):
                j = i + 1
                if firstLoop:
                    range[i] = bounds[i]
                    range[j] = bounds[j]
                    pass
                else:
                    range[i] = min(range[i], bounds[i])
                    range[j] = max(range[j], bounds[j])
                    pass
                pass
            firstLoop = False
            pass

        qxc = 0.5 * (range[0] + range[1])
        qyc = 0.5 * (range[2] + range[3])
        qzc = 0.5 * (range[4] + range[5])

        qxw = 0.5 * (range[1] - range[0]) * (1.0 + margin[0])
        qyw = 0.5 * (range[3] - range[2]) * (1.0 + margin[1])
        qzw = 0.5 * (range[5] - range[4]) * (1.0 + margin[2])

        xrange = GetRoundingRange([qxc - qxw, qxc + qxw], -1)
        yrange = GetRoundingRange([qyc - qyw, qyc + qyw], -1)
        zrange = GetRoundingRange([qzc - qzw, qzc + qzw], -1)

        self.__bounds = (xrange[0], xrange[1],
                         yrange[0], yrange[1],
                         zrange[0], zrange[1])
        return

    def UpdateQpPositions(self, mode):
        if mode == 'DNA':
            QPointClass = DNAQPoint
            pass
        else:
            QPointClass = DGTypeQPoint
            pass

        checkedRegions = self.param.GetAllRegionNames()
        [ids, ide, jds, jde, kds, kde] = self.GetBoundIndices()

        for i in xrange(len(self.__qPointArray)):
            self.__qPointArray[i].SetVisible(False)
            pass

        for i in xrange(ids, ide + 1):
            for j in xrange(jds, jde + 1):
                for k in xrange(kds, kde + 1):
                    indexArray = (i, j, k)
                    [qx, qy, qz] = self.GetRotateQPosition(indexArray)

                    qp = QPointClass(self.param)
                    qp.SetIndex(indexArray[0], indexArray[1], indexArray[2])

                    key = qp.GetIndexString()

                    if key not in self.__qPointHash:
                        self.__qPointArray.append(qp)
                        self.__qPointHash[key] = qp
                        pass
                    else:
                        qp = self.__qPointHash[key]
                        pass

                    qp.SetPosition(qx, qy, qz)
                    qp.CalcConfig(checkedRegions)

                    if (self.__bounds[0] - define.GEOM_TINY_VALUE <= qx <=
                        self.__bounds[1] + define.GEOM_TINY_VALUE and
                        self.__bounds[2] - define.GEOM_TINY_VALUE <= qy <=
                        self.__bounds[3] + define.GEOM_TINY_VALUE and
                        self.__bounds[4] - define.GEOM_TINY_VALUE <= qz <=
                        self.__bounds[5] + define.GEOM_TINY_VALUE):
                        qp.SetVisible(True)
                        pass
                    else:
                        qp.SetVisible(False)
                        pass
                    pass
                pass
            pass
        return

    def GetAllQPoints(self):
        return self.__qPointArray

    def GetVisibleQPoints(self):
        qPoints = []

        for i in xrange(len(self.__qPointArray)):
            if self.__qPointArray[i].IsVisible():
                qPoints.append(self.__qPointArray[i])
            pass

        return qPoints

    def GetQPoint(self, key):
        return self.__qPointHash[key]

    def GetBounds(self):
        return self.__bounds

    def GetBoundingSize(self):
        x = self.__bounds[1] - self.__bounds[0]
        y = self.__bounds[3] - self.__bounds[2]
        z = self.__bounds[5] - self.__bounds[4]
        return (x, y, z)

    def GetCenter(self):
        x = 0.5 * (self.__bounds[1] + self.__bounds[0])
        y = 0.5 * (self.__bounds[3] + self.__bounds[2])
        z = 0.5 * (self.__bounds[5] + self.__bounds[4])
        return (x, y, z)

    def GetDiagonalBoundingSize(self):
        x = self.__bounds[1] - self.__bounds[0]
        y = self.__bounds[3] - self.__bounds[2]
        z = self.__bounds[5] - self.__bounds[4]
        return math.sqrt(x * x + y * y + z * z)

    def ReadDefinition(self, filename):
        self.status.Read(filename)
        return

    def GetRotateQPosition(self, index):
        ra = self.ltc.ra_rotate
        rb = self.ltc.rb_rotate
        rc = self.ltc.rc_rotate

        return vector.GetVectorByIndex(ra, rb, rc, index[0], index[1], index[2])

    def UpdateBoundingBox(self, regions):
        st = self.status

        margin = [st.GetFloat(TAG_GRAPH_AXISMARGIN_X),
                  st.GetFloat(TAG_GRAPH_AXISMARGIN_Y),
                  st.GetFloat(TAG_GRAPH_AXISMARGIN_Z)]

        self.SetBoundingBox(margin, regions)
        return

    def GetScalarFieldRange(self, scalarKind, isMeasurementData):
        smin = 0.0; smax = 0.0

        if scalarKind == define.ENERGY_TRANSFER:
            targetData = self.__detectorRegions

            for name in targetData:
                if targetData[name] == None:
                    continue

                smin = min(np.amin(targetData[name]['etvals']), smin)
                smax = max(np.amax(targetData[name]['etvals']), smax)
                pass

            range = GetRoundingRange((smin, smax), 0)
            pass
        elif scalarKind == define.DETECT_INTENSITY:
            targetData = self.__intensityRegions

            for name in targetData:
                if targetData[name] == None:
                    continue

                smin = min(np.amin(targetData[name]['raw data']), smin)
                smax = max(np.amax(targetData[name]['raw data']), smax)
                pass

            digitNum = math.ceil(math.log10(smax))
            range = GetRoundingRange((smin, smax), digitNum)
            pass

        return range

    def LoadDataFromFile(self, filename, isHDF):
        TOF_BINDATA = 'TOF'
        ENG_BINDATA = 'Energy'

        energyRange = None

        if isHDF:
            import h5py

            HDF5_file = h5py.File(filename, 'r')

            entry_group = HDF5_file['Entry1']
            data_group = entry_group['Data1']
            elmat_group = data_group['ElementContainerMatrixData']
            elary_group = elmat_group['ElementContainerArray0']
            elcon_group = elary_group['ElementContainer0']

            TOF_dataset = elcon_group.get(TOF_BINDATA, None)
            Eng_dataset = elcon_group.get(ENG_BINDATA, None)

            if Eng_dataset != None:
                matrix_header = elmat_group['HeaderBase_in_ElementContainerMatrix']
                ei = matrix_header['Ei'][0]
                etmin = Eng_dataset[0]
                etmax = Eng_dataset[-1]

                energyRange = (ei, etmin, etmax)
                pass
            pass
        else:
            import Manyo

            reader = Manyo.ReadSerializationFileBinary(str(filename))
            elmat_group = reader.LoadElementContainerMatrix()
            elary_group = elmat_group.Put(0)
            elcon_group = elary_group.Put(0)

            if elcon_group.PutXKey() == TOF_BINDATA:
                TOF_dataset = elcon_group.PutX()
                Eng_dataset = None
                pass
            elif elcon_group.PutXKey() == ENG_BINDATA:
                matrix_header = elmat_group.PutHeader()

                TOF_dataset = None
                Eng_dataset = elcon_group.PutX()
                ei = matrix_header.PutDouble('Ei')
                etmin = Eng_dataset[0]
                etmax = Eng_dataset[-1]

                energyRange = (ei, etmin, etmax)
                pass
            pass

        if TOF_dataset != None:
            self.__intensityDataType = TOF_BINDATA
            pass
        elif Eng_dataset != None:
            self.__intensityDataType = ENG_BINDATA
            pass
        else:
            assert 0, 'Illegal nexus file'
            pass

        if self.__intensityDataType == TOF_BINDATA:
            if isHDF:
                self.param.L1 = matrix_header['L1'][0] * MILLI
                TOF_array = np.zeros(shape=TOF_dataset.shape,
                                     dtype=TOF_dataset.dtype)

                TOF_dataset.read_direct(TOF_array)
                pass
            else:
                self.param.L1 = matrix_header.PutDouble('L1') * MILLI
                TOF_array = TOF_dataset 
                pass

            # Calc TOF offset
            BIN_minid = None
            BIN_maxid = None

            vi = DIRAC_CONST / NEUTRON_MASS * self.param.ki / ANGSTROM
            TOF0 = self.param.L1 / vi

            L2max = 0.0

            for props in self.param.detectorProps.values():
                w = max(abs(props.mR - props.nPSD), props.mR) * props.dPSD
                L2x = sqrt(props.L2 ** 2 + w ** 2)

                h = max(abs(props.yd0 - props.nPx * props.hPx), abs(props.yd0))
                L2t = sqrt(L2x ** 2 + h ** 2)
                L2max = max(L2max, L2t)
                pass

            for i in xrange(TOF_array.size):
                vf = L2max / (TOF_array[i] * MICRO - TOF0)
                kf = NEUTRON_MASS / DIRAC_CONST * vf * ANGSTROM

                if kf > 0.0 and self.param.kfmin <= kf <= self.param.kfmax:
                    if BIN_minid == None:
                        BIN_minid = max(i - 1, 0)
                        pass
                    BIN_maxid = i
                    pass
                pass

            if BIN_minid == None:
                if isHDF:
                    HDF5_file.close()
                    pass
                return energyRange

            TOF_array = np.require(TOF_array[BIN_minid:BIN_maxid + 1], dtype = np.float32)
            self.param.TOF_array = TOF_array
            pass
        else:
            if isHDF:
                self.param.nexus_ei = matrix_header['Ei'][0]
                pass
            else:
                self.param.nexus_ei = matrix_header.PutDouble('Ei')
                pass
            self.param.nexus_ki = np.sqrt(self.param.nexus_ei / define.MEV_ENG_FAC)

            if isHDF:
                Eng_array = np.zeros(shape=Eng_dataset.shape,
                                     dtype=Eng_dataset.dtype)

                Eng_dataset.read_direct(Eng_array)
                self.param.Eng_array = np.require(Eng_array, dtype = np.float32)
                pass
            else:
                Eng_array = np.array(Eng_dataset)
                self.param.Eng_array = np.require(Eng_array, dtype = np.float32)
                pass

            BIN_minid = 0; BIN_maxid = Eng_array.size - 1
            pass

        self.__intensityData = dict()

        if isHDF:
            numLoadedDetectors = elmat_group['size'][0]
            pass
        else:
            numLoadedDetectors = elmat_group.PutSize()
            pass

        for i in xrange(numLoadedDetectors):
            if isHDF:
                name = 'ElementContainerArray' + str(i)
                elary_group = elmat_group[name]
                nPx = elary_group['size'][0]
                header = elary_group['HeaderBase_in_ElementContainerArray']
                PSD = header['PSDID'][0]
                pass
            else:
                elary_group = elmat_group.Put(i)
                nPx = elary_group.PutSize()
                header = elary_group.PutHeader()
                PSD = header.PutDouble('PSDID')
                pass

            for region, props in self.param.detectorProps.items():
                locPSD = PSD - props.PSD0
                if 0 <= locPSD < props.nPSD:
                    if self.__intensityData.get(region, None) == None:
                        self.__intensityData[region] = np.zeros(shape=(BIN_maxid - BIN_minid, props.nPx, props.nPSD), dtype = np.float32)
                        pass

                    for Px in xrange(nPx):
                        if isHDF:
                            name = 'ElementContainer' + str(Px)
                            elcon_group = elary_group[name]
                            intensity_dataset = elcon_group['Intensity']
                            intensity_array = np.zeros(shape = intensity_dataset.shape,
                                                         dtype = intensity_dataset.dtype)
                            intensity_dataset.read_direct(intensity_array)
                            pass
                        else:
                            elcon_group = elary_group.Put(Px)
                            intensity_array = elcon_group.PutY()
                            pass
                        intensity_array = np.require(intensity_array, dtype = np.float32)
                        self.__intensityData[region][:, Px, locPSD] = intensity_array[BIN_minid:BIN_maxid]
                        pass
                    break
                pass
            pass

        if isHDF:
            HDF5_file.close()
            pass
        return energyRange

    def MeasurementDataMapping(self, mode):
        for regionName in self.param.GetAllRegionNames():
            if regionName not in self.__intensityData:
                continue

            self.__intensityRegions[regionName] = dict()

            props = self.param.detectorProps[regionName]

            nPx = props.nPx
            nPSD = props.nPSD

            Pxv = np.linspace(0, nPx, nPx + 1)[np.newaxis, :, np.newaxis]
            PSDv = np.linspace(0, nPSD, nPSD + 1)[np.newaxis, np.newaxis, :]

            if mode == 'DNA':
                Etv = self.param.Eng_array[:, np.newaxis, np.newaxis]

                theta_mv = props.theta_m[np.newaxis, :, np.newaxis]
                theta_B = np.arccos(props.y0 / props.RA * np.cos(theta_mv))
                kfv = math.pi / (props.d_miller * np.sin(theta_B))
                Ef = kfv * kfv * define.MEV_ENG_FAC
                Ei = Ef + Etv
                kiv = np.sqrt(Ei / define.MEV_ENG_FAC)

                phi = props.phi_e - props.rPSD * PSDv

                cos_phi = np.cos(phi)
                sin_phi = np.sin(phi)

                cos_theta_mv = np.cos(theta_mv)
                sin_theta_mv = np.sin(theta_mv)

                qz0 = kfv * cos_theta_mv
                qy0 = kfv * sin_theta_mv

                qx = qz0 * sin_phi
                qy = qy0 + np.zeros(PSDv.shape)
                qz = qz0 * cos_phi

                points = np.array([qx + np.zeros(kiv.shape), qy + np.zeros(kiv.shape), qz - kiv])
                etvals = Etv + np.zeros(Pxv.shape) + np.zeros(PSDv.shape)
                pxvals = Pxv + np.zeros(Etv.shape) + np.zeros(PSDv.shape)
                pass
            elif mode == 'SIK':
                w = (props.mR - PSDv) * props.dPSD
                phi = props.rR0 + np.arctan2(w, props.L2)
                L2x = np.sqrt(props.L2 ** 2 + w ** 2)

                cos_phi = np.cos(phi)
                sin_phi = np.sin(phi)

                if self.__intensityDataType == 'TOF':
                    TOFv = self.param.TOF_array[:, np.newaxis, np.newaxis]

                    h = props.yd0 - Pxv * props.hPx
                    L2t = np.sqrt(L2x ** 2 + h ** 2)

                    vi = DIRAC_CONST / NEUTRON_MASS * self.param.ki / ANGSTROM
                    vf = L2t / (TOFv * MICRO - self.param.L1 / vi)
                    kf = NEUTRON_MASS / DIRAC_CONST * vf * ANGSTROM

                    theta_m = np.arctan2(h, L2x)

                    cos_theta_m = np.cos(theta_m)
                    sin_theta_m = np.sin(theta_m)

                    qz0 = kf * cos_theta_m
                    qy0 = kf * sin_theta_m

                    qx = qz0 * sin_phi
                    qy = qy0
                    qz = qz0 * cos_phi - self.param.ki

                    points = np.array([qx, qy, qz])
                    etvals = define.MEV_ENG_FAC * (self.param.ki ** 2 - kf ** 2)
                    pxvals = Pxv + np.zeros(TOFv.shape) + np.zeros(PSDv.shape)
                    pass
                else:
                    Efv = self.param.Eng_array[:, np.newaxis, np.newaxis]

                    kf = np.sqrt((self.param.nexus_ei - Efv) / define.MEV_ENG_FAC)

                    theta_m = np.arctan2(props.yd0 - Pxv * props.hPx, L2x)

                    cos_theta_m = np.cos(theta_m)
                    sin_theta_m = np.sin(theta_m)

                    qz0 = kf * cos_theta_m
                    qy0 = kf * sin_theta_m

                    qx = qz0 * sin_phi
                    qy = qy0
                    qz = qz0 * cos_phi - self.param.nexus_ki

                    points = np.array([qx, qy, qz])
                    etvals = Efv + np.zeros(Pxv.shape) + np.zeros(PSDv.shape)
                    pxvals = Pxv + np.zeros(Efv.shape) + np.zeros(PSDv.shape)
                    pass
            else:
                assert 0, "Illegal Detector Parameter"

            if self.__intensityDataType== 'TOF':
                self.__intensityRegions[regionName]['dimensions'] = (self.param.TOF_array.size, nPx + 1, nPSD + 1)
                gridSize = self.param.TOF_array.size * (nPx + 1) * (nPSD + 1)
                pass
            else:
                self.__intensityRegions[regionName]['dimensions'] = (self.param.Eng_array.size, nPx + 1, nPSD + 1)
                gridSize = self.param.Eng_array.size * (nPx + 1) * (nPSD + 1)
                pass

            self.__intensityRegions[regionName]['raw data'] = self.__intensityData[regionName]
            self.__intensityRegions[regionName]['points'] = points.transpose().reshape(3 * gridSize)
            self.__intensityRegions[regionName]['etvals'] = etvals.transpose().reshape(gridSize)
            self.__intensityRegions[regionName]['pxvals'] = pxvals.transpose().reshape(gridSize)
            pass
        return

    def GetIntensityRegions(self):
        return self.__intensityRegions

def ApplyRotMat(vtkRotMat, pos):
    return vtkRotMat.TransformPoint(pos)

class Lattice(object):
    LT_SYS_TRICLINIC = 'Triclinic'
    LT_SYS_MONOCLINIC = 'Monoclinic'
    LT_SYS_ORTHORHOMBIC = 'Orthorhombic'
    LT_SYS_TETRAGONAL = 'Tetragonal'
    LT_SYS_TRIGONAL = 'Trigonal'
    LT_SYS_HEXAGONAL = 'Hexagonal'
    LT_SYS_CUBIC = 'Cubic'

    LT_TYPE_SIMPLE = 'Simple'
    LT_TYPE_BODY_CENTERED = 'Body-centered'
    LT_TYPE_FACE_CENTERED = 'Face-centered'
    LT_TYPE_BASE_CENTERED = 'Base-centered'
    LT_TYPE_RHOMBOHEDRAL = 'Rhombohedral'

    LATTICE_SYSTEMS = {
        LT_SYS_TRICLINIC:(LT_TYPE_SIMPLE,),
        LT_SYS_MONOCLINIC:(LT_TYPE_SIMPLE, LT_TYPE_BASE_CENTERED,),
        LT_SYS_ORTHORHOMBIC:(LT_TYPE_SIMPLE, LT_TYPE_BASE_CENTERED, LT_TYPE_BODY_CENTERED, LT_TYPE_FACE_CENTERED,),
        LT_SYS_TETRAGONAL:(LT_TYPE_SIMPLE, LT_TYPE_BODY_CENTERED,),
        LT_SYS_TRIGONAL:(LT_TYPE_RHOMBOHEDRAL,),
        LT_SYS_HEXAGONAL:(LT_TYPE_SIMPLE,),
        LT_SYS_CUBIC:(LT_TYPE_SIMPLE, LT_TYPE_BODY_CENTERED, LT_TYPE_FACE_CENTERED,)
        }

    LATTICE_KIND_DENOMS = {
        LT_SYS_TRICLINIC: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ]
            },
        LT_SYS_MONOCLINIC: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ],
            LT_TYPE_BASE_CENTERED: [
                [decimal.Decimal(1), decimal.Decimal(2), None],
                [None, decimal.Decimal(2), None],
                [None, None, decimal.Decimal(1)]
                ]
            },
        LT_SYS_ORTHORHOMBIC: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ],
            LT_TYPE_BASE_CENTERED: [
                [decimal.Decimal(2), decimal.Decimal(-2), None],
                [decimal.Decimal(2), decimal.Decimal(2), None],
                [None, None, 1]
                ],
            LT_TYPE_BODY_CENTERED: [
                [decimal.Decimal(-2), decimal.Decimal(2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(-2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(2), decimal.Decimal(-2)]
                ],
            LT_TYPE_FACE_CENTERED: [
                [None, decimal.Decimal(2), decimal.Decimal(2)],
                [decimal.Decimal(2), None, decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(2), None]
                ],
            },
        LT_SYS_TETRAGONAL: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ],
            LT_TYPE_BODY_CENTERED: [
                [decimal.Decimal(-2), decimal.Decimal(2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(-2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(2), decimal.Decimal(-2)]
                ]
            },
        LT_SYS_TRIGONAL: {
            LT_TYPE_RHOMBOHEDRAL: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ]
            },
        LT_SYS_HEXAGONAL: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ]
            },
        LT_SYS_CUBIC: {
            LT_TYPE_SIMPLE: [
                [decimal.Decimal(1), None, None],
                [None, decimal.Decimal(1), None],
                [None, None, decimal.Decimal(1)]
                ],
            LT_TYPE_BODY_CENTERED: [
                [decimal.Decimal(-2), decimal.Decimal(2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(-2), decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(2), decimal.Decimal(-2)]
                ],
            LT_TYPE_FACE_CENTERED: [
                [None, decimal.Decimal(2), decimal.Decimal(2)],
                [decimal.Decimal(2), None, decimal.Decimal(2)],
                [decimal.Decimal(2), decimal.Decimal(2), None]
                ]
            }
        }

    def __init__(self, status):
        ltcSys = status.GetValue(TAG_SAMPLE_LTCSYS)
        ltcKind = status.GetValue(TAG_SAMPLE_LTCKIND)
        al = status.GetFloat(TAG_SAMPLE_A)
        bl = status.GetFloat(TAG_SAMPLE_B)
        cl = status.GetFloat(TAG_SAMPLE_C)
        alpha = status.GetFloat(TAG_SAMPLE_ALPHA)
        beta = status.GetFloat(TAG_SAMPLE_BETA)
        gamma = status.GetFloat(TAG_SAMPLE_GAMMA)
        uh = status.GetFloat(TAG_SAMPLE_UH)
        uk = status.GetFloat(TAG_SAMPLE_UK)
        ul = status.GetFloat(TAG_SAMPLE_UL)
        vh = status.GetFloat(TAG_SAMPLE_VH)
        vk = status.GetFloat(TAG_SAMPLE_VK)
        vl = status.GetFloat(TAG_SAMPLE_VL)

        self.ltcKindDenom = Lattice.LATTICE_KIND_DENOMS[ltcSys][ltcKind]

        a, b, c = self.GetUnitVector(ltcSys,
                                     al, bl, cl,
                                     alpha, beta, gamma)

        v0 = vector.GetDot(a, vector.GetCross(b, c))
        fac = 2.0 * math.pi / v0

        ra = vector.GetCross(b, c, fac)
        rb = vector.GetCross(c, a, fac)
        rc = vector.GetCross(a, b, fac)

        u = vector.GetVectorByIndex(ra, rb, rc, uh, uk, ul)
        v = vector.GetVectorByIndex(ra, rb, rc, vh, vk, vl)

        # Calc qY-axis rotation angle
        psi = -math.atan2(u[0], u[2])

        ua = [0.0,
              u[1],
              - math.sin(psi) * u[0] + math.cos(psi) * u[2]]

        va = [math.cos(psi) * v[0] + math.sin(psi) * v[2],
              v[1],
              - math.sin(psi) * v[0] + math.cos(psi) * v[2]]

        # Calc qX-axis rotation angle
        chi = math.atan2(ua[1], ua[2])

        vb = [va[0],
              math.cos(chi) * va[1] - math.sin(chi) * va[2],
              0.0]

        # Calc qZ-axis rotation angle
        phi = -math.atan2(vb[1], vb[0])

        R = VTKPanel.GetRotationMatrix(psi, chi, phi)

        self.ra_rotate = self.ra_orig = ApplyRotMat(R, ra)
        self.rb_rotate = self.rb_orig = ApplyRotMat(R, rb)
        self.rc_rotate = self.rc_orig = ApplyRotMat(R, rc)

        self.a_rotate = self.a_orig = ApplyRotMat(R, a)
        self.b_rotate = self.b_orig = ApplyRotMat(R, b)
        self.c_rotate = self.c_orig = ApplyRotMat(R, c)

        self.u = ApplyRotMat(R, u)
        self.v = ApplyRotMat(R, v)

        self.rotUh = uh
        self.rotUk = uk
        self.rotUl = ul
        self.rotVh = vh
        self.rotVk = vk
        self.rotVl = vl
        return

    def GetUnitVector(self,
                      ltcSys,
                      al, bl, cl,
                      alpha, beta, gamma):
        alpha = math.radians(alpha)
        beta = math.radians(beta)
        gamma = math.radians(gamma)

        if ltcSys == Lattice.LT_SYS_TRICLINIC:
            a = [al, 0.0, 0.0]
            b = [bl * math.cos(gamma), bl * math.sin(gamma), 0.0]
            c = [cl * math.cos(alpha),
                 cl * math.cos(beta) * math.sin(gamma),
                 cl * np.sqrt(1.0 - math.cos(alpha) ** 2 + (math.cos(beta) * math.sin(gamma)) ** 2)]
            pass
        elif ltcSys == Lattice.LT_SYS_MONOCLINIC:
            a = [al, 0.0, 0.0]
            b = [0.0, bl, 0.0]
            c = [cl * math.cos(beta), 0.0, cl * math.sin(beta)]
            pass
        elif ltcSys == Lattice.LT_SYS_ORTHORHOMBIC:
            a = [al, 0.0, 0.0]
            b = [0.0, bl, 0.0]
            c = [0.0, 0.0, cl]
            pass
        elif ltcSys == Lattice.LT_SYS_TETRAGONAL:
            a = [al, 0.0, 0.0]
            b = [0.0, al, 0.0]
            c = [0.0, 0.0, cl]
            pass
        elif ltcSys == Lattice.LT_SYS_TRIGONAL:
            s = 2.0 * al / np.sqrt(3.0) * math.sin(0.5 * alpha)
            r = np.sqrt(al * al - s * s)

            a = [0.0, s, r]
            b = [0.5 * s * np.sqrt(3.0), -0.5 * s, r]
            c = [-0.5 * s * np.sqrt(3.0), -0.5 * s, r]
            pass
        elif ltcSys == Lattice.LT_SYS_HEXAGONAL:
            a = [0.5 * np.sqrt(3.0) * al, -0.5 * al, 0.0]
            b = [0.0, al, 0.0]
            c = [0.0, 0.0, cl]
            pass
        elif ltcSys == Lattice.LT_SYS_CUBIC:
            a = [al, 0.0, 0.0]
            b = [0.0, al, 0.0]
            c = [0.0, 0.0, al]
            pass
        return a, b, c

    def SetRotation(self, psi, chi, phi):
        R = VTKPanel.GetRotationMatrix(psi, chi, phi)

        self.ra_rotate = ApplyRotMat(R, self.ra_orig)
        self.rb_rotate = ApplyRotMat(R, self.rb_orig)
        self.rc_rotate = ApplyRotMat(R, self.rc_orig)

        self.a_rotate = ApplyRotMat(R, self.a_orig)
        self.b_rotate = ApplyRotMat(R, self.b_orig)
        self.c_rotate = ApplyRotMat(R, self.c_orig)

        fac = 1.0 / (2.0 * math.pi)

        self.rotUh = vector.GetDot(self.a_rotate, self.u) * fac
        self.rotUk = vector.GetDot(self.b_rotate, self.u) * fac
        self.rotUl = vector.GetDot(self.c_rotate, self.u) * fac
        self.rotVh = vector.GetDot(self.a_rotate, self.v) * fac
        self.rotVk = vector.GetDot(self.b_rotate, self.v) * fac
        self.rotVl = vector.GetDot(self.c_rotate, self.v) * fac

        if abs(self.rotUh) < define.GEOM_TINY_VALUE:
            self.rotUh = 0.0
            pass

        if abs(self.rotUk) < define.GEOM_TINY_VALUE:
            self.rotUk = 0.0
            pass

        if abs(self.rotUl) < define.GEOM_TINY_VALUE:
            self.rotUl = 0.0
            pass

        if abs(self.rotVh) < define.GEOM_TINY_VALUE:
            self.rotVh = 0.0
            pass

        if abs(self.rotVk) < define.GEOM_TINY_VALUE:
            self.rotVk = 0.0
            pass

        if abs(self.rotVl) < define.GEOM_TINY_VALUE:
            self.rotVl = 0.0
            pass
        return

try:
    import elementtree.ElementTree as ET # for Python 2.4
    pass
except ImportError:
    import xml.etree.ElementTree as ET # for Python 2.5 or later.
    pass

XML = ET.XML
ElementTree = ET.ElementTree
Element = ET.Element
iselement = ET.iselement
SubElement = ET.SubElement

class StatusError:
    def __init__(self, message):
        self.message = message
        return

    def __str__(self):
        return repr('Status Error: ' + self.message)

    def GetMessage(self):
        return self.message

# TAG NAMES FOR CONFIGULATIONS
TAG_ROOT = ('MethodAssistant',)

TAG_ROTATE = TAG_ROOT + ('rotate',)
TAG_ROTATE_PSI = TAG_ROTATE + ('psi',)
TAG_ROTATE_CHI = TAG_ROTATE + ('chi',)
TAG_ROTATE_PHI = TAG_ROTATE + ('phi',)
TAG_ROTATE_GUIDEAXIS = TAG_ROTATE + ('guide_axis',)

TAG_GRAPH = TAG_ROOT + ('graph_settings',)
TAG_GRAPH_AXISMARGIN = TAG_GRAPH + ('axis_margin',)
TAG_GRAPH_AXISMARGIN_X = TAG_GRAPH_AXISMARGIN + ('x',)
TAG_GRAPH_AXISMARGIN_Y = TAG_GRAPH_AXISMARGIN + ('y',)
TAG_GRAPH_AXISMARGIN_Z = TAG_GRAPH_AXISMARGIN + ('z',)

TAG_GRAPH_AXISDISPLAY = TAG_GRAPH + ('axis_display',)
TAG_GRAPH_AXISDISPLAY_X = TAG_GRAPH_AXISDISPLAY + ('x',)
TAG_GRAPH_AXISDISPLAY_Y = TAG_GRAPH_AXISDISPLAY + ('y',)
TAG_GRAPH_AXISDISPLAY_Z = TAG_GRAPH_AXISDISPLAY + ('z',)

TAG_OBJ3D = TAG_ROOT + ('object_settings',)
TAG_OBJ3D_ETVAL = TAG_OBJ3D + ('etval',)
TAG_OBJ3D_ETWIDTH = TAG_OBJ3D + ('etwidth',)
TAG_OBJ3D_QPCOLOR = TAG_OBJ3D + ('qp_color',)
TAG_OBJ3D_QPDISPLAY = TAG_OBJ3D + ('qp_display_mode',)
TAG_OBJ3D_SURFSCALAR = TAG_OBJ3D + ('surface_scalar',)
TAG_OBJ3D_PXNO = TAG_OBJ3D + ('pixel_no',)
TAG_OBJ3D_PXRANGE = TAG_OBJ3D + ('pixel_range',)
TAG_OBJ3D_ETRANGE = TAG_OBJ3D + ('et_range',)
TAG_OBJ3D_ETRANGE_MODE = TAG_OBJ3D_ETRANGE + ('mode',)
TAG_OBJ3D_ETRANGE_MIN = TAG_OBJ3D_ETRANGE + ('min',)
TAG_OBJ3D_ETRANGE_MAX = TAG_OBJ3D_ETRANGE + ('max',)
TAG_OBJ3D_INTENSITYRANGE = TAG_OBJ3D + ('intensity_range',)
TAG_OBJ3D_INTENSITYRANGE_MODE = TAG_OBJ3D_INTENSITYRANGE + ('mode',)
TAG_OBJ3D_INTENSITYRANGE_MIN = TAG_OBJ3D_INTENSITYRANGE + ('min',)
TAG_OBJ3D_INTENSITYRANGE_MAX = TAG_OBJ3D_INTENSITYRANGE + ('max',)

TAG_ENG = TAG_ROOT + ('energy',)
TAG_ENG_EIMIN = TAG_ENG + ('minimum_incident_energy',)
TAG_ENG_EIMAX = TAG_ENG + ('maximum_incident_energy',)
TAG_ENG_EI = TAG_ENG + ('incident_energy',)
TAG_ENG_ETMIN = TAG_ENG + ('minimum_transition_energy',)
TAG_ENG_ETMAX = TAG_ENG + ('maximum_transition_energy',)

TAG_QEGRAPH = TAG_ROOT + ('QE_graph_settings',)

TAG_QEGRAPH_PLANEDEF = TAG_QEGRAPH + ('plane_definition',)
TAG_QEGRAPH_OH = TAG_QEGRAPH + ('Oh',)
TAG_QEGRAPH_OK = TAG_QEGRAPH + ('Ok',)
TAG_QEGRAPH_OL = TAG_QEGRAPH + ('Ol',)
TAG_QEGRAPH_UH = TAG_QEGRAPH + ('Uh',)
TAG_QEGRAPH_UK = TAG_QEGRAPH + ('Uk',)
TAG_QEGRAPH_UL = TAG_QEGRAPH + ('Ul',)
TAG_QEGRAPH_VH = TAG_QEGRAPH + ('Vh',)
TAG_QEGRAPH_VK = TAG_QEGRAPH + ('Vk',)
TAG_QEGRAPH_VL = TAG_QEGRAPH + ('Vl',)

TAG_QRGRAPH = TAG_ROOT + ('QR_graph_settings',)

TAG_QRGRAPH_PLANEDEF = TAG_QRGRAPH + ('plane_definition',)
TAG_QRGRAPH_OH = TAG_QRGRAPH + ('Oh',)
TAG_QRGRAPH_OK = TAG_QRGRAPH + ('Ok',)
TAG_QRGRAPH_OL = TAG_QRGRAPH + ('Ol',)
TAG_QRGRAPH_UH = TAG_QRGRAPH + ('Uh',)
TAG_QRGRAPH_UK = TAG_QRGRAPH + ('Uk',)
TAG_QRGRAPH_UL = TAG_QRGRAPH + ('Ul',)
TAG_QRGRAPH_VH = TAG_QRGRAPH + ('Vh',)
TAG_QRGRAPH_VK = TAG_QRGRAPH + ('Vk',)
TAG_QRGRAPH_VL = TAG_QRGRAPH + ('Vl',)
TAG_QRGRAPH_ROTATEAXIS = TAG_QRGRAPH + ('rotate_axis',)
TAG_QRGRAPH_NSTEP = TAG_QRGRAPH + ('nstep',)
TAG_QRGRAPH_INTERVAL = TAG_QRGRAPH + ('interval',)
TAG_QRGRAPH_PSI0 = TAG_QRGRAPH + ('psi0',)
TAG_QRGRAPH_CHI0 = TAG_QRGRAPH + ('chi0',)
TAG_QRGRAPH_PHI0 = TAG_QRGRAPH + ('phi0',)
TAG_QRGRAPH_ETMIN = TAG_QRGRAPH + ('minimum_transition_energy',)
TAG_QRGRAPH_ETMAX = TAG_QRGRAPH + ('maximum_transition_energy',)

TAG_SAMPLE = TAG_ROOT + ('sample',)

TAG_SAMPLE_NAME = TAG_SAMPLE + ('name',)
TAG_SAMPLE_LTCSYS = TAG_SAMPLE + ('lattice_system',)
TAG_SAMPLE_LTCKIND = TAG_SAMPLE + ('lattice_kind',)
TAG_SAMPLE_A = TAG_SAMPLE + ('a',)
TAG_SAMPLE_B = TAG_SAMPLE + ('b',)
TAG_SAMPLE_C = TAG_SAMPLE + ('c',)
TAG_SAMPLE_ALPHA = TAG_SAMPLE + ('alpha',)
TAG_SAMPLE_BETA = TAG_SAMPLE + ('beta',)
TAG_SAMPLE_GAMMA = TAG_SAMPLE + ('gamma',)
TAG_SAMPLE_UH = TAG_SAMPLE + ('Uh',)
TAG_SAMPLE_UK = TAG_SAMPLE + ('Uk',)
TAG_SAMPLE_UL = TAG_SAMPLE + ('Ul',)
TAG_SAMPLE_VH = TAG_SAMPLE + ('Vh',)
TAG_SAMPLE_VK = TAG_SAMPLE + ('Vk',)
TAG_SAMPLE_VL = TAG_SAMPLE + ('Vl',)

TAG_SELECTQPINDEX = TAG_ROOT + ('select_Q_point_index',)

class Status(object):
    def __init__(self):
        self._data = dict()
        self._tree = ElementTree()
        return

    def SetValue(self, key, value):
        prev_value = self._data.get(key, None)
        has_key = self._data.has_key(key)

        self._data[key] = value
        return

    def GetValue(self, key):
        return self._data.get(key, None)

    def GetFloat(self, key):
        value = self.GetValue(key)

        if isinstance(value, float):
            return value
        elif isinstance(value, decimal.Decimal):
            return float(value)

        return None

    def ParseXml(self, source):
        self._tree.parse(source)
        return

    def OutputStatus(self, filepath):
        # Copy data to elementtree
        for key, value in self._data.items():
            path = self._GetXMLPath(key)
            elem = self._tree.find(path)

            if iselement(elem):
                elem.text = str(value)
                pass
            pass

        # Output xml data
        self._tree.write(filepath, 'utf-8')
        return

    def ReadValue(self, key, trans = str):
        path = self._GetXMLPath(key)
        elem = self._tree.find(path)

        if iselement(elem):
            try:
                value = trans(elem.text)
                pass
            except:
                raise StatusError('Cannot convert xml data of ' + path + ' to ' + str(trans) + '.')
            else:
                self.SetValue(key, value)
                pass
            pass
        else:
            raise StatusError('Cannot find xml data of ' + path)
        return

    def _GetXMLPath(self, key):
        if len(key) <= 1:
            return '/'

        first = True
        for name in key:
            if first:
                path = '.'
                first = False
                pass
            else:
                path += '/' + name
                pass
            pass
        return path

    def Read(self, filepath):
        self.ParseXml(filepath)

        self.ReadValue(TAG_ROTATE_PSI, decimal.Decimal)
        self.ReadValue(TAG_ROTATE_CHI, decimal.Decimal)
        self.ReadValue(TAG_ROTATE_PHI, decimal.Decimal)
        self.ReadValue(TAG_ROTATE_GUIDEAXIS, str)

        self.ReadValue(TAG_GRAPH_AXISMARGIN_X, decimal.Decimal)
        self.ReadValue(TAG_GRAPH_AXISMARGIN_Y, decimal.Decimal)
        self.ReadValue(TAG_GRAPH_AXISMARGIN_Z, decimal.Decimal)
        
        self.ReadValue(TAG_GRAPH_AXISDISPLAY_X, bool)
        self.ReadValue(TAG_GRAPH_AXISDISPLAY_Y, bool)
        self.ReadValue(TAG_GRAPH_AXISDISPLAY_Z, bool)
        
        self.ReadValue(TAG_OBJ3D_ETVAL, decimal.Decimal)
        self.ReadValue(TAG_OBJ3D_ETWIDTH, decimal.Decimal)

        self.ReadValue(TAG_OBJ3D_QPCOLOR, int)
        self.ReadValue(TAG_OBJ3D_QPDISPLAY, int)
        self.ReadValue(TAG_OBJ3D_SURFSCALAR, str)
        self.ReadValue(TAG_OBJ3D_PXNO, int)
        self.ReadValue(TAG_OBJ3D_PXRANGE, int)
        self.ReadValue(TAG_OBJ3D_ETRANGE_MODE, str)
        self.ReadValue(TAG_OBJ3D_ETRANGE_MIN, decimal.Decimal)
        self.ReadValue(TAG_OBJ3D_ETRANGE_MAX, decimal.Decimal)
        self.ReadValue(TAG_OBJ3D_INTENSITYRANGE_MODE, str)
        self.ReadValue(TAG_OBJ3D_INTENSITYRANGE_MIN, decimal.Decimal)
        self.ReadValue(TAG_OBJ3D_INTENSITYRANGE_MAX, decimal.Decimal)

        self.ReadValue(TAG_ENG_EIMIN, decimal.Decimal) # for DNA
        self.ReadValue(TAG_ENG_EIMAX, decimal.Decimal) # for DNA
        self.ReadValue(TAG_ENG_EI, decimal.Decimal)    # for SIK, AMR
        self.ReadValue(TAG_ENG_ETMIN, decimal.Decimal) # for SIK, AMR
        self.ReadValue(TAG_ENG_ETMAX, decimal.Decimal) # for SIK, AMR

        self.ReadValue(TAG_QEGRAPH_PLANEDEF, int)
        self.ReadValue(TAG_QEGRAPH_OH, int)
        self.ReadValue(TAG_QEGRAPH_OK, int)
        self.ReadValue(TAG_QEGRAPH_OL, int)
        self.ReadValue(TAG_QEGRAPH_UH, int)
        self.ReadValue(TAG_QEGRAPH_UK, int)
        self.ReadValue(TAG_QEGRAPH_UL, int)
        self.ReadValue(TAG_QEGRAPH_VH, int)
        self.ReadValue(TAG_QEGRAPH_VK, int)
        self.ReadValue(TAG_QEGRAPH_VL, int)

        self.ReadValue(TAG_QRGRAPH_PLANEDEF, int)
        self.ReadValue(TAG_QRGRAPH_OH, int)
        self.ReadValue(TAG_QRGRAPH_OK, int)
        self.ReadValue(TAG_QRGRAPH_OL, int)
        self.ReadValue(TAG_QRGRAPH_UH, int)
        self.ReadValue(TAG_QRGRAPH_UK, int)
        self.ReadValue(TAG_QRGRAPH_UL, int)
        self.ReadValue(TAG_QRGRAPH_VH, int)
        self.ReadValue(TAG_QRGRAPH_VK, int)
        self.ReadValue(TAG_QRGRAPH_VL, int)
        self.ReadValue(TAG_QRGRAPH_PSI0, decimal.Decimal)
        self.ReadValue(TAG_QRGRAPH_CHI0, decimal.Decimal)
        self.ReadValue(TAG_QRGRAPH_PHI0, decimal.Decimal)
        self.ReadValue(TAG_QRGRAPH_ROTATEAXIS, int)
        self.ReadValue(TAG_QRGRAPH_NSTEP, int)
        self.ReadValue(TAG_QRGRAPH_INTERVAL, decimal.Decimal)
        self.ReadValue(TAG_QRGRAPH_ETMIN, decimal.Decimal)
        self.ReadValue(TAG_QRGRAPH_ETMAX, decimal.Decimal)

        self.ReadValue(TAG_SAMPLE_NAME, str)
        self.ReadValue(TAG_SAMPLE_LTCSYS)
        self.ReadValue(TAG_SAMPLE_LTCKIND)
        self.ReadValue(TAG_SAMPLE_A, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_B, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_C, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_ALPHA, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_BETA, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_GAMMA, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_UH, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_UK, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_UL, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_VH, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_VK, decimal.Decimal)
        self.ReadValue(TAG_SAMPLE_VL, decimal.Decimal)

        self.SetValue(TAG_SELECTQPINDEX, None)
        return

class AssistantConfError:
    def __init__(self, message):
        self.message = message
        return

    def __str__(self):
        return repr('AssistantConfError: ' + self.message)

    def GetMessage(self):
        return self.message

ASSISTANT_CONF_FILEPATH = u'./AssistantConfs.xml'

class AssistantConfs(object):
    def __init__(self, inst, filename = ASSISTANT_CONF_FILEPATH):
        import os

        filepath = os.path.join(os.path.abspath(os.path.dirname(__file__)), filename)

        self.etree = ElementTree()
        self.etree.parse(filepath)

        self.conf = None

        elems = self.etree.findall('./AssistantConf')

        for elem in elems:
            value = elem.get('inst', None)

            if value == None:
                raise AssistantConfError('Cannot find attribute (inst) for tag (AssistantConf) in file (', +filepath + ').')
            elif value == inst:
                self.conf = elem
                break
            pass

        if self.conf == None:
            raise AssistantConfError('Cannot find tag (AssistantConf) or appropriate attribute (inst) for (' + inst + ') in file (', +filepath + ')')
        return

    def GetRegions(self):
        regions = self.conf.find('Regions')
        if regions == None:
            raise AssistantConfError('Cannot find tag (Regions)')
        return regions

def GetElemValue(elem, tag, trans):
    text = elem.findtext(tag, None)
    if text == None:
        raise AssistantConfError('Cannot find tag (' + tag + ')')

    try:
        value = trans(text)
        return value
    except:
        raise AssistantConfError('Find illegal value(' + text + ') for tag (' + tag + ')')
    return

def GetWaveNumberByEnergy(E):
    return math.sqrt(2.0 * define.NEUTRON_MASS * E * define.MILLI_ELECTRON_VOLT) / define.DIRAC_CONST * define.ANGSTROM

class DetectorParamsBase(object):
    def GetAllRegionNames(self):
        items = self.detectorProps.items()
        items.sort(lambda a, b:cmp(a[1].PSD0, b[1].PSD0))

        names = []
        for item in items:
            names.append(item[0])
            pass
        return names

class DNADetectorParams(DetectorParamsBase):
    def __init__(self, elemRegions):
        self.detectorProps = dict()

        elems = elemRegions.findall('Region')

        for elem in elems:
            label = elem.get('label')
            self.detectorProps[label] = DNARegionParams(elem)
            pass
        return

    def SetEnergyRange(self, eimin, eimax):
        self.kimin = GetWaveNumberByEnergy(eimin)
        self.kimax = GetWaveNumberByEnergy(eimax)
        return

# NEWTON METHOD PARAMETERS
NEWTON_MAX_ITERS = 100
NETWON_CRITERIA = 1.0e-12

class DNARegionParams(object):
    def __init__(self, elemRegion):
        d_miller = GetElemValue(elemRegion, 'd_miller', float)
        Rd = GetElemValue(elemRegion, 'Rd', float)
        RA = GetElemValue(elemRegion, 'RA', float)
        y0 = GetElemValue(elemRegion, 'y0', float)
        rR0 = GetElemValue(elemRegion, 'rR0', float)
        dPSD = GetElemValue(elemRegion, 'dPSD', float)
        nPSD = GetElemValue(elemRegion, 'nPSD', int)
        mR = GetElemValue(elemRegion, 'mR', float)
        PSD0 = GetElemValue(elemRegion, 'PSD0', int)
        LPSD = GetElemValue(elemRegion, 'LPSD', float)
        nPx = GetElemValue(elemRegion, 'nPx', int)
        yd0 = GetElemValue(elemRegion, 'yd0', float)
        rdpP = 2

        Rd = define.MILLI * Rd
        RA = define.MILLI * RA
        y0 = define.MILLI * y0
        rR0 = math.radians(rR0)
        dPSD = define.MILLI * dPSD
        LPSD = define.MILLI * LPSD
        yd0 = define.MILLI * yd0
        rPSD = 2.0 * math.atan2(0.5 * dPSD, Rd)
        mR = mR - 0.5
        hPx = LPSD / nPx

        self.d_miller = d_miller
        self.Rd = Rd
        self.RA = RA
        self.y0 = y0
        self.rPSD = rPSD
        self.nPSD = nPSD
        self.mR = mR
        self.PSD0 = PSD0
        self.hPx = hPx
        self.nPx = nPx
        self.yd0 = yd0
        self.rdpP = rdpP

        assert nPx % rdpP == 0, "rdpP must be one of the measure of nP."

        self.phi_s = rR0 + rPSD * (mR - nPSD)
        self.phi_e = rR0 + rPSD * mR

        # Calculate pixel angles

        self.theta_m = [0.0] * (nPx + 1)
        r = y0 / RA

        for Px in xrange(nPx + 1):
            if Px == 0:
                theta_m = 0.0
                pass
            else:
                theta_m = self.theta_m[Px - 1]
                pass

            yP = yd0 - hPx * Px

            for iter in xrange(NEWTON_MAX_ITERS):
                sin_theta_m = math.sin(theta_m)
                cos_theta_m = math.cos(theta_m)

                theta_B = math.acos(r * cos_theta_m)

                cos_theta_B = r * cos_theta_m
                sin_theta_B = math.sin(theta_B)

                Lsa = RA * (sin_theta_B + r * sin_theta_m)

                y0m = Lsa * sin_theta_m
                z0m = Lsa * cos_theta_m

                theta_X = math.asin((y0 - y0m) / RA)

                yd = y0m + (z0m - Rd) / math.tan(theta_B - theta_X)

                dThetaBByTheta_m = r * sin_theta_m / sin_theta_B

                dLsaByTheta_m = RA * cos_theta_B * (1.0 + dThetaBByTheta_m)

                dY0mByTheta_m = Lsa * cos_theta_m + sin_theta_m * dLsaByTheta_m
                dZ0mByTheta_m = -Lsa * sin_theta_m + cos_theta_m * dLsaByTheta_m

                dThetaXByTheta_m = -dY0mByTheta_m / math.sqrt(RA ** 2 - (y0 - y0m) ** 2)

                dYdByTheta_m = dY0mByTheta_m + dZ0mByTheta_m / math.tan(theta_B - theta_X) - \
                    (z0m - Rd) / math.sin(theta_B - theta_X) ** 2 * \
                    (dThetaBByTheta_m - dThetaXByTheta_m)

                dTheta_m = -(yd - yP) / dYdByTheta_m
                theta_m += dTheta_m

                if (max(abs(dTheta_m), abs(yd - yP)) < NETWON_CRITERIA):
                    break;
                pass
            self.theta_m[Px] = theta_m
            pass

        self.theta_m = np.array(self.theta_m)
        self.theta_m_max = max(self.theta_m[0], self.theta_m[self.nPx])
        self.theta_m_min = min(self.theta_m[0], self.theta_m[self.nPx])
        return

class DGTypeDetectorParam(DetectorParamsBase):
    def __init__(self, elemRegions):
        self.detectorProps = dict()

        elems = elemRegions.findall('Region')

        for elem in elems:
            label = elem.get('label')
            self.detectorProps[label] = DGTypeRegionParams(elem)
            pass
        return

    def SetEnergyRange(self, ei, etmin, etmax):
        self.ki = GetWaveNumberByEnergy(ei)
        self.kfmax = GetWaveNumberByEnergy(ei - etmin)
        self.kfmin = GetWaveNumberByEnergy(ei - etmax)
        return

class DGTypeRegionParams(object):
    def __init__(self, elemRegion):
        rR0 = GetElemValue(elemRegion, 'rR0', float)
        nPSD = GetElemValue(elemRegion, 'nPSD', int)
        PSD0 = GetElemValue(elemRegion, 'PSD0', int)
        mR = GetElemValue(elemRegion, 'mR', float)
        dPSD = GetElemValue(elemRegion, 'dPSD', float)
        yd0 = GetElemValue(elemRegion, 'yd0', float)
        LPSD = GetElemValue(elemRegion, 'LPSD', float)
        nPx = GetElemValue(elemRegion, 'nPx', int)
        L2 = GetElemValue(elemRegion, 'L2', float)
        rdpP = 1

        rR0 = math.radians(rR0)
        mR = mR - 0.5
        dPSD = dPSD * define.MILLI
        LPSD = LPSD * define.MILLI
        yd0 = yd0 * define.MILLI
        L2 = L2 * define.MILLI
        hPx = LPSD / nPx

        self.rR0 = rR0
        self.nPSD = nPSD
        self.PSD0 = PSD0
        self.mR = mR
        self.dPSD = dPSD
        self.hPx = hPx
        self.nPx = nPx
        self.yd0 = yd0
        self.L2 = L2

        self.rdpP = rdpP

        assert nPx % rdpP == 0, "rdpP must be one of the measure of nP."

        self.theta_m_min_0 = math.atan2(yd0 - LPSD, L2)
        self.theta_m_max_0 = math.atan2(yd0, L2)

        self.phi_s = rR0 + math.atan2(dPSD * (mR - nPSD), L2)
        self.phi_e = rR0 + math.atan2(dPSD * mR, L2)
        return



from __future__ import print_function

import Manyo.SAS as ms
import Manyo as mm
import Manyo.Utsusemi as mu
import os
import math
"""
Base Functions for SAS
QSpace.py
- ConvertQSpace
- QzProjection
- SectorAverageByAngles
- MakeRegionOfSectAve
"""
from utsusemi.SAS.ana.Reduction.SASUtils import SASBankDic
########################################
def ConvertQSpace( dat, InVec="0.0 0.0 1.0" ):
    """
    Convert Data to Q space EventData to produce histogram
    ConvertQSpace(dat=DAT, InVec="0.0 0.0 1.0")

    @param  dat    (ElementContainerMatrix)
    @param  InVec  (string) vector of the incident beam
    @retval None
    """
    in_vec_st = InVec.split(" ")
    if (len(in_vec_st)!=3):
        raise UserWarning("InVec parameter is invalid.")
    vx = float( in_vec_st[0] )
    vy = float( in_vec_st[1] )
    vz = float( in_vec_st[2] )

    CT = ms.ConvertToQspace( dat )
    CT.SetIncidentBeamVector( vx, vy, vz )
    CT.Execute()
    del CT

########################################
def ConvertQSpaceRPMTCenterShift(dat, DET, PIX):
    """
    Convert Data to Qspace on RPMT using shifted center position.
    The center position is at center point of the pixel given by DET and PIX.
    If DET or PIX is minus, for example -1, then this function calculates with the original position.

    ConvertQSpaceRPMTCenterShift(dat, DET, PIX)

    @param dat (ElementContainerMatrix) Target Data
    @param DET (int) Detect ID, or minus value (-1) to use original position
    @param PIX (int) Pixel No, or minus value (-1) to use original position
    @retval None
    """
    CT = ms.ConvertToQspace(dat)
    if CT.SetIncidentBeamRPMT(DET, PIX):
        CT.Execute()
    del CT

########################################
def ConvertQSpaceWithGravity( dat, params="188.19 0.1562 208.5" ):
    """
    Convert Data to Q space EventData to produce histogram with gravity effect
    ConvertQSpaceWithGravity(dat=DAT, params="188.19 0.1562 208.5")
    py = a+b*(lambda)^2

    params = "188.19 0.1562 208.5" <- a, b, px
    params = "0.1562 1.0 208.5 188.35" <- b, lambda, px, py@lambda

    @param  dat    (ElementContainerMatrix)
    @param  params (string) "a b px" or "b lambda px py"
    @retval None
    """
    param_list_st_sp = params.split(" ")
    param_list_st_cm = params.split(",")
    if len(param_list_st_sp) > len(param_list_st_cm):
        param_list_st = param_list_st_sp[:]
    else:
        param_list_st = param_list_st_cm[:]

    CT = ms.ConvertToQspace( dat )
    if (len(param_list_st)==1):
        modeNo = 5
        runNo_st = dat.PutHeader().PutString(mu.UTSUSEMI_KEY_HEAD_RUNNUMBER)
        runNo_sv = runNo_st.split(",")
        runNo = int(runNo_sv[0])
        CT.SetIncidentBeamWithGravityRPMTDetectInfo(runNo, modeNo)
    elif (len(param_list_st)==3):
        CT.SetIncidentBeamWithGravity( float(param_list_st[0]), float(param_list_st[1]), float(param_list_st[2]) )
    elif (len(param_list_st)==4):
        CT.SetIncidentBeamWithGravity( float(param_list_st[0]), float(param_list_st[1]), float(param_list_st[2]), float(param_list_st[3]) )
    else:
        raise UserWarning("params is invalid.")

    CT.Execute()
    del CT

########################################
def QzProjection( dat, QxRange="-1.0 1.0 0.005", QyRange="-1.0 1.0 0.005", QzRange="-0.5 0.0", LambdaRange="0.7 7.6", Bank="All" ):
    """
    Makes Qx-Qy map data integrated in a Qz range.
    QzProjection( dat=DAT, QxRange="-1.0 1.0 0.005", QyRange="-1.0 1.0 0.005", QzRange="-0.5 0.0" )

    @param  dat   (ElementContainerMatrix) Data after ConvertToQSpace
    @param  QxRange   (string)  "Qx_min Qx_max Qx_bin"
    @param  QyRange   (string)  "Qy_min Qy_max Qy_bin"
    @param  QzRange   (string)  "Qz_max Qz_bin"
    @param LambdaRange(string)  "Lambda_min, Lambda_max"
    @param  Bank      (string)  "All" or "<BankName> <BankName> ...", delimiter can be " ", "," or "-"
    @retval ECA   (ElementContainerArray)
    """

    qx_range_st=QxRange.split(" ")
    qy_range_st=QyRange.split(" ")
    qz_range_st=QzRange.split(" ")
    lmd_range_st=LambdaRange.split(" ")

    qxmin = float( qx_range_st[0] )
    qxmax = float( qx_range_st[1] )
    qxbin = float( qx_range_st[2] )

    qymin = float( qy_range_st[0] )
    qymax = float( qy_range_st[1] )
    qybin = float( qy_range_st[2] )

    qzmin = float( qz_range_st[0] )
    qzmax = float( qz_range_st[1] )

    lmdmin = float( lmd_range_st[0] )
    lmdmax = float( lmd_range_st[1] )

    bank_id_list = []
    if Bank.upper() != "ALL":
        delim_list = [len(Bank.split(" ")), len(Bank.split(",")), len(Bank.split("-"))]
        delim_ind = delim_list.index(max(delim_list))
        delim = " "
        if delim_ind == 1:
            delim = ","
        elif delim_ind == 2:
            delim = "-"
        bank_split_sp = Bank.split(delim)
        for a_bank in bank_split_sp:
            if a_bank=="SM":
                bank_id_list.append(SASBankDic["SMBL14"])
                bank_id_list.append(SASBankDic["SMBL16"])
            elif a_bank in SASBankDic:
                bank_id_list.append(SASBankDic[a_bank])
            else:
                UserWarning("QzProjection >> Invalid Bank name {}".format(a_bank))

    QP = ms.QzProjection( dat )
    QP.SetMapRange( qxmin, qxmax, qxbin, qymin, qymax, qybin )
    QP.SetQzRange( qzmin, qzmax )
    QP.SetLambRange( lmdmin, lmdmax )
    for a_bankId in bank_id_list:
        QP.AddBankId(a_bankId)
    QP.Execute()

    ret_eca = mm.ElementContainerArray()
    QP.SetResultAsECA( ret_eca )

    del QP
    try:
        hh = ret_eca.PutHeaderPointer()
        #hh.AddHeader("YLABEL", "Qy (1/Ang.)")
        #hh.AddHeader("XLABEL", "Qx (1/Ang.)")
        hh.Add("YLABEL", "Qy")
        hh.Add("XLABEL", "Qx")
        hh.Add("Xunit", "1/Ang.")
        hh.Add("Yunit", "1/Ang.")
        hh.Add("M2PP_X_KEY","Qx")
        hh_dat = dat.PutHeaderPointer()
        if hh_dat.CheckKey("TransmittanceParameter"):
            hh.AddHeader("TransmittanceParameter", hh_dat.PutDoubleVector("TransmittanceParameter"))
        if hh_dat.CheckKey("TransmittanceParameterError"):
            hh.AddHeader("TransmittanceParameterError", hh_dat.PutDoubleVector("TransmittanceParameterError"))
        if hh_dat.CheckKey("TransmittanceFitFunc"):
            hh.AddHeader("TransmittanceFitFunc", hh_dat.PutString("TransmittanceFitFunc"))
        if hh_dat.CheckKey("RUNNUMBER"):
            hh.AddHeader("RUNNUMBER", hh_dat.PutString("RUNNUMBER"))
    except:
        pass
    return ret_eca

########################################
def SectorAverageByAngles( dat, Banks="SM16-SM14-SD-SU", Qrange="0.01 2.0 0.001 dQ/Q", LamRange="0.7 0.8", AzimAng="0.0 0.0", PolAng="0.0 0.0", QLamFile="None/QLambdaRange" ):
    """
    Convert Data to Q space EventData to produce histogram
    SectorAverageByAngles( dat=DAT, Banks="SM16-SM14-SD-SU", Qrange="0.01 2.0 0.001 dq/Q", LamRange="0.7 0.8", AzimAng="0.0 0.0", PolAng="0.0 0.0" )

    To give Q binning, default format is "<Qmin> <Qmax> <DeltaQ> <Type>". Another method is given by Text file described list of Q values.
    Format :
        - The line with "#Type=" at the head is parameter for Q binning type { dq, dq/q, None }
        - The line with "#" at the head is ignored

    Example 1 : Free q value list ( Descibe Q binning with a Q value for each line )
        #QbinType{dq,dq/q,None}
        #Type=None
        0.01
        0.02
        0.05
        ...
    Example 2 : delta-Q const   ( Add dQ value after Q value with comma )
        #QbinType{dq,dq/q,None}
        #Type=dq
        0.01,0.01
        0.5,0.02
        1.0,0.04
        ...
    Example 3 : delta-Q/Q const   ( Add dQ/Q value after Q value with comma )
        #QbinType{dq,dq/q,None}
        #Type=dq/q
        0.01,0.01
        0.5,0.02
        1.0,0.04
        ...

    @param  dat  (ElementContainerMatrix)
    @param  Banks    (string)  "SM16-SD-SU" means these banks are unified in one Iq, "SM SD SU" means separated Iq
    @param  Qrange   (string)  Q range and type of Qbin "Qmin Qmax deltaQ type" type must be dQ/Q or dQ. Or Qrange.txt
    @param  LamRange (string)  Lambda range "Lambda_min Lambda_max"
    @param  AzimAng  (string)  "-20.0 20 170 -170"
    @param  PolAng   (string)  "0 90 0 40"
    @param  QLamFile (string)  filename of Q-Lambda range info. If None, no use. (read QLambdaRange_SM.txt for SM bank)
    @retval Iq       (list of ElementContainer)
    """

    Banks = Banks.strip().upper()

    isUnifiedIq = False
    bankId_list = []
    bankName_list = []
    if len(Banks.split(" "))!=1:
        tmp_bank_list_s = Banks.split(" ")
        for bb in tmp_bank_list_s:
            if bb=="SM16":
                bb="SMBL16"
            if bb=="SM14":
                bb="SMBL14"
            if bb=="SM":
                bankId_list.append( SASBankDic["SMBL16"] )
                bankId_list.append( SASBankDic["SMBL14"] )
                bankName_list.append( "SMBL16" )
                bankName_list.append( "SMBL14" )
            elif bb in SASBankDic:
                bankId_list.append( SASBankDic[bb] )
                bankName_list.append( bb )
        if len(bankId_list)==0:
            msg = "Cmm.SectorAverageByAngles > Banks is invalid ("+Banks+")"
            raise UserWarning(msg)
    elif len(Banks.split("-"))!=1:
        tmp_bank_list_s = Banks.split("-")
        for bb in tmp_bank_list_s:
            if bb=="SM16":
                bb="SMBL16"
            if bb=="SM14":
                bb="SMBL14"
            if bb=="SM":
                bankId_list.append( SASBankDic["SMBL16"] )
                bankId_list.append( SASBankDic["SMBL14"] )
                bankName_list.append( "SMBL16" )
                bankName_list.append( "SMBL14" )
            elif bb in SASBankDic:
                bankId_list.append( SASBankDic[bb] )
                bankName_list.append( bb )
        if len(bankId_list)==0:
            msg = "Cmm.SectorAverageByAngles > Banks is invalid ("+Banks+")"
            raise UserWarning(msg)
        isUnifiedIq = True
    else:
        if Banks=="SM16":
            Banks="SMBL16"
        if Banks=="SM14":
            Banks="SMBL14"

        if Banks=="SM":
            bankId_list.append( SASBankDic["SMBL16"] )
            bankId_list.append( SASBankDic["SMBL14"] )
            bankName_list.append( "SMBL16" )
            bankName_list.append( "SMBL14" )
            isUnifiedIq = True
        elif Banks=="ALL":
            bankId_list = list(SASBankDic.values())
            bankName_list = list(SASBankDic.keys())
        elif Banks in SASBankDic:
            bankId_list.append( SASBankDic[Banks] )
            bankName_list.append( Banks )
        else:
            msg = "Cmm.SectorAverageByAngles > Banks is invalid ("+Banks+")"
            raise UserWarning(msg)

    QbinType = 0
    QRlist = []
    QbinInfo = []
    QbinList = []
    qr_list_st = Qrange.split(" ")
    if len(qr_list_st)==1:
        p_file=mu.FindParamFilePath( qr_list_st[0] )
        if p_file=="":
            raise UserWarning(" Qrange is invalid")
        fo = open( p_file, "r" )
        p_conts_list = fo.readlines()
        fo.close()
        if (len(p_conts_list)==0):
            raise UserWarning(" Qrange File is invalid ("+p_file+")")
        ind_lines = 0
        while(True):
            if p_conts_list[ind_lines].find("#Type=")==0:
                qbin_type=p_conts_list[ind_lines][6:].strip()
                if qbin_type.lower()=="dq":
                    QbinType=0
                elif qbin_type.lower()=="dq/q":
                    QbinType=1
            elif p_conts_list[ind_lines].find("#")!=0:
                break
            ind_lines += 1
        split_char = ":"
        tmp_a_line_sp = p_conts_list[ind_lines].split(split_char)
        if len(tmp_a_line_sp)==1:
            tmp_a_line_sp2 = p_conts_list[ind_lines].split(",")
            if len(tmp_a_line_sp2)!=1:
                split_char = ","
                tmp_a_line_sp = tmp_a_line_sp2[:]

        if len(tmp_a_line_sp)==1: ## Free Hist
            QbinList.append( float(p_conts_list[ind_lines].strip()) )
            ind_lines += 1
            while( len(p_conts_list)>ind_lines ):
                if p_conts_list[ind_lines].find("#")!=0:
                    if (p_conts_list[ind_lines].strip())!="":
                        QbinList.append( float(p_conts_list[ind_lines].strip()) )
                ind_lines +=1

        elif len(tmp_a_line_sp)==2: ## Qbin with bin info
            QbinInfo.append( float( tmp_a_line_sp[0].strip() ) )
            QbinInfo.append( float( tmp_a_line_sp[1].strip() ) )
            ind_lines += 1
            while( len(p_conts_list)>ind_lines ):
                if p_conts_list[ind_lines].find("#")!=0:
                    tmp_a_line_sp = p_conts_list[ind_lines].split(split_char)
                    if len(tmp_a_line_sp)>=2:
                        QbinInfo.append( float( tmp_a_line_sp[0].strip() ) )
                        QbinInfo.append( float( tmp_a_line_sp[1].strip() ) )
                    elif len(tmp_a_line_sp)==1: # Last line
                        if tmp_a_line_sp[0].strip()!="":
                            QbinInfo.append( float( tmp_a_line_sp[0].strip() ) )
                            QbinInfo.append( 0.0 )
                            break
                ind_lines += 1
        else:
            raise UserWarning(" Qrange File is invalid.  ("+p_file+")")

    elif len(qr_list_st)==3:
        for i in range(3):
            QRlist.append( float(qr_list_st[i]) )
    elif len(qr_list_st)==4:
        if qr_list_st[3].strip().lower()=="dq/q":
            QbinType=1
        for i in range(3):
            QRlist.append( float(qr_list_st[i]) )
    else:
        raise UserWarning(" Qrange is invalid."+Qrange)
    """
    if len(qr_list_st)!=3 and len(qr_list_st)!=4:
        raise UserWarning, " Qrange is invalid."+Qrange

    qr_list = [ float(qr_list_st[0]), float(qr_list_st[1]), float(qr_list_st[2]) ]
    QbinType = 0
    if len(qr_list_st)==4:
        if (qr_list_st[3].lower()=="dq/q"):
            QbinType = 1
    """

    lam_list_st = LamRange.split(" ")
    lam_list = []
    for lam in lam_list_st:
        if lam!="":
            lam_list.append( float(lam) )
    if len(lam_list)!=2:
        raise UserWarning(" LamRange is invalid.")

    az_list_st = AzimAng.split(" ")
    az_list = []
    for az in az_list_st:
        if az!="":
            az_list.append( float(az) )
    if len(az_list)==2:
        if (az_list[0]==0.0) and (az_list[1]==0.0):
            az_list=[]

    po_list_st = PolAng.split(" ")
    po_list = []
    for po in po_list_st:
        if po!="":
            po_list.append( float(po) )

    v_angles = mm.MakeDoubleVector()
    if len(az_list)!=0:
        if len(az_list)==len(po_list):

            for i in range( 0,len(az_list),2 ):
                v_angles.append( az_list[i] )
                v_angles.append( az_list[i+1] )
                v_angles.append( po_list[i] )
                v_angles.append( po_list[i+1] )
        else:
            raise UserWarning("Angle Info are invalid")

    ## QLamFile check
    QLam_range_bank = []
    if QLamFile[:4].lower()!="none":
        for bank_name in bankName_list:
            qlam_range = []
            qlam_file = "%s_%s.txt" % (QLamFile, bank_name)
            #qlam_file_path = os.path.join( os.environ["UTSUSEMI_USR_DIR"],"ana","xml",qlam_file )
            qlam_file_path = mu.FindParamFilePath( qlam_file )
            if not os.path.exists(qlam_file_path):
                raise UserWarning("file given by QLamFile argument is not existed. name="+qlam_file)
            fo = open( qlam_file_path, "r" )
            rline = fo.readline()
            while( rline!="" ):
                if rline[0]!="#":
                    lamline_st = rline.split(",")
                    if len(lamline_st)==3:
                        qlam_range.append( (float(lamline_st[0]), float(lamline_st[1]), float(lamline_st[2]) ) )
                rline = fo.readline()
            fo.close()
            QLam_range_bank.append( (SASBankDic[bank_name], qlam_range[:]) )

    ## [counts/lambda] -> [counts] instead of Jacobian calc
    #tt = ms.ConvertToPoint(dat)
    #tt.Execute(True)
    #del tt
    ## Executes sector-averaging
    cc=ms.SectorAverageByAngle(dat)
    if (isUnifiedIq):
        bank_v = mm.MakeInt4Vector()
        for bb in bankId_list:
            bank_v.append( bb )
        cc.SetTargetBank( bank_v )

    if len(QLam_range_bank)!=0:
        for (bank_id,QLam_range) in QLam_range_bank:
            for (lam,qmin,qmax) in QLam_range:
                cc.AddQLambdaRangeToIq( bank_id, lam, qmin, qmax )

    if len(QRlist)!=0:
        cc.SetQRange( QbinType, QRlist[0], QRlist[1], QRlist[2] ) ## Set Q range for I(Q)
    elif len(QbinInfo)!=0:
        QbinInfo_v = mm.MakeDoubleVector()
        for a_qbin in QbinInfo:
            QbinInfo_v.append(a_qbin)
        cc.SetQRange( QbinType, QbinInfo_v ) ## Set Q range for I(Q)
    elif len(QbinList)!=0:
        QbinList_v = mm.MakeDoubleVector()
        for a_qbin in QbinList:
            QbinList_v.append(a_qbin)
        cc.SetQRange( QbinList_v ) ## Set Q range for I(Q)

    cc.SetLambdaRange( lam_list[0], lam_list[1] )  ## Set Lambda range
    if v_angles.size()!=0:
        cc.SetSectAngles( v_angles )
    cc.Execute()

    ret = []
    if (isUnifiedIq):
        print( "SectorAverage.py : bankId_list[0]=",bankId_list[0])
        a_ec = cc.PutIq( bankId_list[0] )
        if (a_ec.PutSize()!=0):
            ret.append( a_ec )
    else:
        for i in bankId_list:
            print( "SectorAverage.py : bankId_list[i]=",i)
            a_ec = cc.PutIq( i )
            if (a_ec.PutSize()!=0):
                ret.append( a_ec )
    del cc

    ## [counts]->[counts/lambda] instead of Jacobian calc
    #tt = ms.ConvertToPoint(dat)
    #tt.Execute()
    #del tt

    """
    ss=""
    for i in bankId_list:
        ss+="%02d,"%(i)
    print "@@@@ BankID_list ="+ss
    if (isUnifiedIq):
        print "@@@@ Unified Iq"
    else:
        print "@@@@ Separated Iqs"
    """

    """
    num_of_iq=cc.PutNumOfIq()
    ret = []
    for i in range(num_of_iq):
        ret.append( cc.PutIq(i) ) ## get I(Q)
    """
    """
    ret = []
    if SASBankDic.has_key(Bank):
        if Bank=="SM":
            ret.append( cc.PutIq( SASBankDic["SMBL16"] ) )
            ret.append( cc.PutIq( SASBankDic["SMBL14"] ) )
        else:
            ret.append( cc.PutIq( SASBankDic[Bank] ) )
        del cc

    else:
        del cc
        raise UserWarning, "Bank name is invalid"
    """
    try:
        hh_dat = dat.PutHeaderPointer()
        for ret_ec in ret:
            hh = ret_ec.PutHeaderPointer()
            if hh_dat.CheckKey("TransmittanceParameter"):
                hh.AddHeader("TransmittanceParameter", hh_dat.PutDoubleVector("TransmittanceParameter"))
            if hh_dat.CheckKey("TransmittanceParameterError"):
                hh.AddHeader("TransmittanceParameterError", hh_dat.PutDoubleVector("TransmittanceParameterError"))
            if hh_dat.CheckKey("TransmittanceFitFunc"):
                hh.AddHeader("TransmittanceFitFunc", hh_dat.PutString("TransmittanceFitFunc"))
    except:
        pass
    return ret

########################################
def MakeRegionOfSectAve( dat, Bank="SM", Qrange="0.01 2.0 0.001 dQ/Q", LamRange="0.7 7.6 0.1", AzimAng="0.0 0.0", PolAng="0.0 0.0", ErrTh="1.0 e", OutFile="QLambdaRange", TwoThetaRange="0.0-180.0" ):
    """
    Makes Lamda slice SectorAverage
    MakeRegionOfSectAve( dat=DAT, Bank="SM", Qrange="0.01 2.0 0.001 dQ/Q", LamRange="0.7 7.6 0.1", AzimAng="0.0 0.0", PolAng="0.0 0.0", ErrTh=1.0, OutFile="QLambdaRange" )

    @param  dat  (ElementContainerMatrix)
    @param  Bank     (string)  Target bank.
    @param  Qrange   (string)  Q range and type of Qbin "Qmin Qmax deltaQ type" type must be dQ/Q or dQ
    @param  LamRange (string)  Lambda range "Lambda_min Lambda_max"
    @param  AzimAng  (string)  "-20.0 20 170 -170"
    @param  PolAng   (string)  "0 90 0 40"
    @param  ErrTh    (string)   "sh type" sh : threshold of error bar to decide useless region (ErrTh<=0.0 means no decision), type="e" or "ie"
    @param  OutFile  (string)  output file name for Q-Lambda range for target bunk (filename = QLambdaRange_SM.xml)
    @param  TwoThetaRange (string)
    @retval IQLam    (ElementContainerArray)
    """

    lam_list_st = LamRange.split(" ")
    if len(lam_list_st)!=3:
        raise UserWarning(" Lamrange is invalid.")
    lam_min = abs( float( lam_list_st[0] ) )
    lam_max = abs( float( lam_list_st[1] ) )
    lam_bin = abs( float( lam_list_st[2] ) )

    if (lam_min>lam_max):
        lam_tmp = lam_min
        lam_min=lam_max
        lam_max = lam_tmp

    lam_cur = lam_min - 0.5*lam_bin
    lam_list = []
    while(lam_cur<=lam_max):
        lam_list.append( (lam_cur,(lam_cur+lam_bin)) )
        lam_cur += lam_bin

    hh = mm.HeaderBase()
    xlabel = "Lamb [%4.2f+ i*%4.2f]"%(lam_min,lam_bin)
    hh.Add("XLABEL",xlabel)
    hh.Add("YLABEL","Q [1/Ang]")
    IQlambda = mm.ElementContainerArray(hh)
    print(" Slice along Lamgda")
    for lamRng in lam_list:
        lamRng_st = "%f %f" % lamRng
        print("lamda range= "+lamRng_st)
        Iq_res = SectorAverageByAngles( dat, Bank, Qrange, lamRng_st, AzimAng, PolAng)
        if len(Iq_res)!=1:
            raise UserWarning("Bank argument must be only one bank")
        IQlambda.Add( Iq_res[0] )

    err_th_st = ErrTh.split(" ")
    if len(err_th_st)!=2:
        raise UserWarning("ErrTh argument is invalid")
    err_th = float( err_th_st[0] )
    th_type = err_th_st[1].lower()
    if not th_type in ["e","ie"]:
        raise UserWarning("ErrTh argument is invalid")

    print("MakeRegionOfSectAve >> err_th=",err_th)
    print("MakeRegionOfSectAve >> th_type=",th_type)

    UCV = mu.UtsusemiUnitConverter()

    # Check TwoThetaRange argument
    _TwoThetaRange = []
    two_theta_list = TwoThetaRange.split("-")
    if len(two_theta_list)==2:
        _twotheta_min = float( two_theta_list[0] )
        _twotheta_max = float( two_theta_list[1] )
        if _twotheta_min!=0.0 or _twotheta_max!=180.0:
            _TwoThetaRange = [ _twotheta_min, _twotheta_max ]
            print("MakeRegionOfSectAve >> Two Theta Range = [%g, %g]"%(_TwoThetaRange[0],_TwoThetaRange[1]))
            #err_th=-1.0

    run_number_st = dat.PutHeaderPointer().PutString("RUNNUMBER")

    ss="## Q range limitation for each wave length lambda\n"
    ss+="## Run Number=%s\n" % (run_number_st)
    ss+="## Bank = %s\n" % (Bank)
    ss+="##\n"
    ss+="## lambda[A],Qmin[1/A],Qmax[1/A]\n"
    #print(ss)

    for i in range(len(lam_list)):
        iqlambda = IQlambda(i)
        lamrange = lam_list[i]
        lamb = (lamrange[0]+lamrange[1])/2.0

        ii=iqlambda.PutY()
        #ee=iqlambda.PutE()
        ee=iqlambda.PutP("Error")
        qq=iqlambda.Put("Q")
        ## Set MASK to bin by given threshold
        if (err_th>0.0):
            for j in range(ii.size()):
                if th_type=="e":
                    if (abs(ee[j])>err_th):
                        ee[j]=-1.0*(abs(ee[j]))
                else:
                    if (ii[j]!=0.0):
                        if (abs(ee[j]/ii[j])>err_th):
                            ee[j]=-1.0*(abs(ee[j]))
                    else:
                        ee[j]=-1.0*(abs(ee[j]))
        if (len(_TwoThetaRange)==2):
            for j in range(ii.size()):
                #(*it_kval) = (UU->VtoLambda())/(*it_lamb)*1000.0 * (UU->Vmm_msToK());
                kf = (UCV.VtoLambda())/lamb*1000.0*UCV.Vmm_msToK()
                sin_val = (qq[j]/2.0)/kf
                two_theta = -1.0
                if sin_val<1.0:
                    two_theta_rad = 2.0*math.asin( sin_val )
                    two_theta = two_theta_rad/math.pi*180.0
                if two_theta==-1.0 or two_theta<_TwoThetaRange[0] or two_theta>_TwoThetaRange[1] :
                    ee[j] = -1.0*(abs(ee[j]))

        qmin=-1.0
        qmax=-1.0
        ii_max = ii.size()-1
        for j in range(1,ii.size()):
            ##print "#### %f/%f, %f/%f / %f/%f, %f/%f" %(ee[j-1],ii[j-1],ee[j],ii[j],ee[ii_max-(j-1)],ii[ii_max-(j-1)],ee[ii_max-j],ii[ii_max-j])
            #if (qmin<0.0) and (ee[j-1]<0.0) and (ee[j]>=0.0):
            if (qmin<0.0) and (ee[j-1]<=0.0) and (ee[j]>0.0):
                qmin = qq[j]
            #if (qmax<0.0) and (ee[ii_max-(j-1)]<0.0) and (ee[ii_max-j]>=0.0):
            if (qmax<0.0) and (ee[ii_max-(j-1)]<=0.0) and (ee[ii_max-j]>0.0):
                qmax = qq[ii_max-j]
            if (qmin!=-1.0) and (qmax!=-1.0):
                break
        if (qmin==-1.0):
            print("--Qmin Search")
            for j in range(ii.size()):
                #print "##search Qmin ",qq[j],ii[j],ee[j]
                if ii[j]>0.0:
                    qmin = qq[j]
                    break
        if (qmax==-1.0):
            print("--Qmax Search")
            for j in range(ii.size()):
                #print "##search Qmax : ",qq[ii_max-j],ii[ii_max-j],ee[ii_max-j]
                if ii[ii_max-j]>0.0:
                    qmax = qq[ii_max-j]
                    break



        #print("%f, %f, %f" %(lamb,qmin,qmax))
        ss += "%f, %f, %f\n" % (lamb,qmin,qmax)

    bank_name_list = []
    if Bank=="SM":
        bank_name_list.append("SMBL16")
        bank_name_list.append("SMBL14")
    else:
        bank_name_list.append(Bank)

    for bank_name in bank_name_list:
        fname = "%s_%s.txt"%(OutFile,bank_name)
        #fpath = os.path.join( os.environ["UTSUSEMI_USR_DIR"],"ana","xml", fname )
        fpath = mu.FindParamFilePath( fname )
        if fpath!="":
            mu.UtsusemiWarning("MakeRegionOfSectAve >> Overwrite %s"%(fpath))
        else:
            if mu.UTSUSEMIMANYOVERSION<"4.0.0":
                fpath = os.path.join( os.environ["UTSUSEMI_USR_DIR"],"ana","xml", fname )
            else:
                fpath = os.path.join( os.environ["UTSUSEMI_BASE_DIR"],os.environ["UTSUSEMI_INST_CODE"],"ana","xml", fname )

        fo = open( fpath, "w" )
        fo.write( ss )
        fo.close()


    #return IQlambda
    xx_new = IQlambda(0).PutXList()
    num_of_lambda = IQlambda.PutSize()

    matrix = []
    matrix_e = []
    for i in range(len(xx_new)-1):
        a_line = [ 0.0 for i in range(num_of_lambda) ]
        matrix.append( a_line[:] )
        matrix_e.append( a_line[:] )

    for i in range(num_of_lambda):
        ec = IQlambda(i)
        yy_ec = ec.PutYList()
        ee_ec = ec.PutEList()
        for j,ax in enumerate(yy_ec):
            if ee_ec[j]>0.0:
                matrix[j][i] = ax
                matrix_e[j][i] = ee_ec[j]

    XX = []
    for ax1,ax2 in lam_list:
        XX.append(ax1)
    XX.append( lam_list[-1][1] )

    hh = mm.HeaderBase()
    #xlabel = "Q [1/Ang]"
    #hh.Add("XLABEL",xlabel)
    #hh.Add("YLABEL","Lambda")
    hh.Add("XLABEL","Q")
    hh.Add("Xunit", "1/Ang")
    hh.Add("YLABEL","Lambda")
    hh.Add("Yunit", "Ang")
    hh.Add("M2PP_X_KEY","Q")
    ret = mm.ElementContainerArray(hh)
    for i in range( len(xx_new)-1 ):
        xrange_v = mm.MakeDoubleVector()
        xrange_v.append( xx_new[i] )
        xrange_v.append( xx_new[i+1] )
        header = mm.HeaderBase()
        header.Add( "XRANGE", xrange_v )
        ec = mm.ElementContainer(header)
        ec.Add( "X", XX )
        ec.Add( "Y", matrix[i] )
        ec.Add( "E", matrix_e[i] )
        ec.SetKeys("X","Y","E")
        ret.Add(ec)
    return ret

########################################
def GetAzimuthProfile(dat, Qcenter=0.036, dQ=0.002, azim_bin=1.0, azim_range="0.0 360.0", lambda_range="0 0"):
    """
    Calculates Azimuth Profile from DataReduction

    @param dat          (ElementContainerMatrix)
    @param Qcenter      (float)
    @param dQ           (float)
    @param azim_bin     (float)
    @param azim_range   (string)
    @param lambda_range (string)
    @retval Iap
    """
    ec_list=GetAzimuthProfileMulti([dat], Qcenter, dQ, azim_bin, azim_range, lambda_range)
    return ec_list[0]

########################################
def GetAzimuthProfileMulti(dat_list, Qcenter=0.036, dQ=0.002, azim_bin=1.0, azim_range="0.0 360.0", lambda_range="0 0"):
    """
    Calculates Azimuth Profile from DataReduction

    @param dat_list     (list)
    @param Qcenter      (float)
    @param dQ           (float)
    @param azim_bin     (float)
    @param azim_range   (string)
    @param lambda_range (string)
    @retval IapList
    """
    azim_range_list = azim_range.split(" ")
    azim_min = float(azim_range_list[0])
    azim_max = float(azim_range_list[1])

    ec_list = []
    CAP = ms.CalcAzimuthProfile()
    if CAP.SetQRadius( Qcenter, dQ ):
        if CAP.SetAzimuthBinning( azim_bin, azim_min, azim_max ):
            if lambda_range!="0 0":
                l_list = lambda_range.split(" ")
                CAP.SetLambRange( float( l_list[0] ), float( l_list[1] ) )

            for i,dat in enumerate(dat_list):
                if not isinstance(dat, mm.ElementContainerMatrix):
                    raise UserWarning("Given data is not ElementContainerMatrix")
                p_chk = mu.UtsusemiCheckDataProcess()
                if not p_chk.CheckProcess(dat,"ConvertToQspace"):
                    raise UserWarning("Given data is invalid ( must do ConvertToQspace before )")
                CAP.SetTarget( dat )
                EC = CAP.GetProfileAverage()
                hh = EC.PutHeaderPointer()
                hh.Add( "AZIMUTHPROFILE","%03dfrm_Q%g_dQ%g"%(i+1,Qcenter,dQ) )
                ec_list.append( EC )
        else:
            raise UserWarning("Azimuth parameter is invalid.")
    else:
        raise UserWarning("Q or dQ parameter is invalid.")

    del CAP
    return ec_list



# -*- coding: utf-8 -*-
#
# $Date: 2012-07-23 19:56:22 +0900 (月, 23 7 2012) $
# $Rev: 17 $
#
import math

def GetNorm(x):
    return math.sqrt(x[0] * x[0] + x[1] * x[1] + x[2] * x[2])

def GetMultiplyVector(x, scalar):
    return [x[0] * scalar, x[1] * scalar, x[2] * scalar]

def GetCross(x, y, fac=1.0):
    return [(x[1] * y[2] - x[2] * y[1]) * fac,
            (x[2] * y[0] - x[0] * y[2]) * fac,
            (x[0] * y[1] - x[1] * y[0]) * fac]

def GetNormalizedVector(x):
    norm = GetNorm(x)

    y = list(x)

    if norm > GEOM_TINY_VALUE:
        y[0] /= norm
        y[1] /= norm
        y[2] /= norm
        return y
    else:
        return None

def GetVectorByIndex(a, b, c, i, j, k):
    return [i * a[0] + j * b[0] + k * c[0],
            i * a[1] + j * b[1] + k * c[1],
            i * a[2] + j * b[2] + k * c[2]]

def GetDot(a, b):
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2]
